# import libraries
import json
from typing import Any, List, Optional
import json
from time import time
from datetime  import timedelta, datetime
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature





class UserPlayTime(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._cumulative_play_time : Optional[timedelta] = None 
        self._cumulative_active_time : Optional[timedelta] = None
        


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["PlayTime"] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
    

        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
    
        # Exception handling for empty sessions
        if(not self._cumulative_play_time):
            self._cumulative_play_time = feature._vals[0]
        if(not self._cumulative_active_time):
            self._cumulative_active_time = feature._vals[1]

        try:
            self._cumulative_play_time  += feature.FeatureValues[0]
            self._cumulative_active_time += feature.FeatureValues[1]
        except:
            self._cumulative_play_time += timedelta(0)
            self._cumulative_active_time += timedelta(0)

        
        
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """

        return [self._cumulative_play_time, self._cumulative_active_time]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Active"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



# import libraries
from typing import Any, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from datetime import date, datetime, time
# import local files
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class SessionStart(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._date : Optional[date] = None
        self._time : Optional[time] = None
        self._year : Optional[int] = None
        self._month : Optional[int] = None
        self._hour : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        event_time = event.Timestamp
        if not self._date:
            self._date = event_time.date()
            self._time = event_time.time()
            self._hour = event_time.hour
            self._year = event_time.year
            self._month = event_time.month


        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._date, self._time, self._year, self._month, self._hour]

    def Subfeatures(self) -> List[str]:
        return ["Time", "Year", "Month", "Hour"]

    def BaseFeatureSuffix(self) -> str:
        return "Date"




class LabCompleteCount:
    def __init__(self):
        self.complete_count = 0

    def track_event(self, event_data):
        if event_data is not None and event_data.get("percent_complete", 0) >= 100:
            self.complete_count += 1

    def get_count(self):
        return self.complete_count



from typing import List, Optional, Dict, Any, Optional, Tuple, Callable
from collections import defaultdict
from models.SequenceModel import SequenceModel
import datetime
import operator
from functools import partial

# popular lakeland event lists

# _ACTIVE_EVENTS = ["SELECTTILE", "SELECTFARMBIT", "SELECTITEM", "SELECTBUY", "BUY",
#                   "CANCELBUY", "TILEUSESELECT", "ITEMUSESELECT", 'STARTGAME', 'TOGGLENUTRITION',
#                   "TOGGLESHOP", "TOGGLEACHIEVEMENTS", "SKIPTUTORIAL", "SPEED"]
# [3, 4, 5, 6, 7, 8, 10, 11, 1, 12, 13, 14, 15, 16]

# _EXPLORATORY_EVENTS = ['SELECTTILE', 'SELECTFARMBIT', 'SELECTITEM', 'TOGGLEACHIEVEMENTS', 'TOGGLESHOP',
#                        'TOGGLENUTRITION']
# [3, 4, 5, 14, 13, 12]


# _IMPACT_EVENTS = ["BUY", "TILEUSESELECT", "ITEMUSESELECT"]
# [7, 10, 11]

## @class TimeSinceEventTypesModel
# Returns time (in seconds) since last received an event from event_list, defaulting to time since first event if
# the given events have never been received.
# @param levels: Levels applicable for model
# @param event_list: events since which time is calculated (using any event, not all events)
# @param filters: A map from event enums to a list of lists. Each sublist represents a boolean function. Let f be a sublist.
# f[0] = event enum (no need to be distinct; a single enum can have multiple filters)
# f[1...-3] = keys and subkeys of an event
# f[-2] = "lt", "le", "eq", "ne", "ge", "gt", "in", or "notin"
# f[-1] = comparator value

class TimeSinceEventTypesModel(SequenceModel):
    def __init__(self, levels: List[int] = [], event_list: List[Any] = [0]):
        '''
        @class TimeSinceEventTypesModel
        Returns time (in seconds) since last received an event from event_list, defaulting to time since first event if
        the given events have never been received.
        :param levels: Levels applicable for model
        :param event_list: Either:
        List of integer events since which time is calculated (using any event, not all events) and
        lists of enum followed by filter arguments.
        Filter arguments are a list of lists. Each sublist represents a boolean function. Let f be a sublist.
        f[3n] = keys and subkeys of an event
        f[3n+1] = "lt", "le", "eq", "ne", "ge", "gt", or "in"
        f[3n+2] = comparator value
        Within a sublist, functions ALL have to return true. Outside a sublist, ANY can return true.

        Example:
        event_list = [
            1,
            2,
            [ 3, [
            [["event_data_complex", "buy"], "eq", 2],
            [["event_data_complex", "worth"], ">", 400]
            ]],
            [ 3 [
            [["event_data_complex", "buy"], "eq", 6],
            [["event_data_complex", "type"], "notin", ["farm", "livestock"]]
            ]]
        ]
        Would return the time since an event matching:
        event_custom==1 OR
        event_custom==2 OR
        (event_custom==3 AND buy==2 AND worth>400) OR
        (event_custom==3 AND buy==6 AND type not in ["farm", "livestock"])
        and defaults to the first event if none found.

        Example:
      "event_list": [
        [7, [
          [["event_data_complex", "buy"], "in", [1,3,5]],
          [["event_data_complex", "success"], "eq", true]
        ]]
      ]
        Would match any event_custom==7 (buy) where buy==1 (house), buy==3 (farm), or buy==5 (dairy), and the buy was a
        success. It would return the time since the most recent matching event in seconds.

        '''

        self._event_list = event_list
        self._parsed_filters = defaultdict(lambda: [])
        for event in event_list:
            if type(event) is int: # no filter
                self._parsed_filters[event].append(lambda event: True)
            else: # should be a 2-el list where the second element is a list of bool func lists
                event_enum, bool_funcs_all = event
                parsed_filter = partial(_base_filter, funcs_all=bool_funcs_all)
                self._parsed_filters[event_enum].append(parsed_filter)

        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> Optional[float]:
        if not events:
            return None
        now = events[-1]["client_time"] # assume this script in the same timezone as server, and server exports
        for event in reversed(events):
            event_filters = self._parsed_filters[event["event_custom"]]
            # if event_filters:
            #     print(event)
            if any(f(event) for f in event_filters):
                break

        event_time = event["client_time"]
        if type(event_time) is str: # fix for tests, datetimes don't always get parsed ahead of time
            event_time = datetime.datetime.fromisoformat(event_time)
        if type(now) is str:
            now = datetime.datetime.fromisoformat(now)
        return (now - event_time).total_seconds()


    def __repr__(self):
        return f"TimeSinceEventTypesModel(event_list={self._event_list},"\
               f", levels={self._levels}, input_type={self._input_type})"


_comparison_str_to_func = {
    "lt": operator.lt,
    "<": operator.lt,
    "le": operator.le,
    "<=": operator.le,
    "eq": operator.eq,
    "==": operator.eq,
    "ne": operator.ne,
    "!=": operator.ne,
    "ge": operator.ge,
    ">=": operator.ge,
    "gt": operator.gt,
    ">": operator.gt,
    "in": lambda a,b: a in b,
    "notin": lambda a, b: a not in b,
}


def _base_filter(event, funcs_all):
    for keys, comparison_str, value in funcs_all:
        event_value = event
        for key in keys:
            # print(event_value, key)
            # print(type(event_value))
            event_value = event_value[key]
        comparison_func = _comparison_str_to_func[comparison_str]
        if not comparison_func(event_value, value):
            return False
    return True

def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]






## import standard libraries
import json
import logging
import re
from dateutil import parser
from datetime import datetime, time, timedelta
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Final, List, Tuple, Optional, Union
## import local files
from ogd.core import schemas
from ogd.core.models.Event import Event, EventSource
from ogd.core.schemas.tables.ColumnMapSchema import ColumnMapSchema
from ogd.core.schemas.tables.ColumnSchema import ColumnSchema
from ogd.core.utils import utils
from ogd.core.utils.Logger import Logger
from ogd.core.utils.typing import Map

## @class TableSchema
#  Dumb struct to hold useful info about the structure of database data
#  for a particular game.
#  This includes the indices of several important database columns, the names
#  of the database columns, the max and min levels in the game, and a list of
#  IDs for the game sessions in the given requested date range.
class TableSchema:

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, schema_name:str, schema_path:Path = Path(schemas.__file__).parent / "table_schemas/"):
        """Constructor for the TableSchema class.
        Given a database connection and a game data request,
        this retrieves a bit of information from the database to fill in the
        class variables.

        :param schema_name: The filename for the table schema JSON.
        :type schema_name: str
        :param schema_path: Path to find the given table schema file, defaults to "./schemas/table_schemas/"
        :type schema_path: str, optional
        :param is_legacy: [description], defaults to False
        :type is_legacy: bool, optional
        """
        # declare and initialize vars
        self._schema            : Optional[Dict[str, Any]]
        self._column_map        : ColumnMapSchema
        self._columns           : List[ColumnSchema] = []
        self._table_format_name : str                    = schema_name

        if not self._table_format_name.lower().endswith(".json"):
            self._table_format_name += ".json"
        self._schema = utils.loadJSONFile(filename=self._table_format_name, path=schema_path)

        # after loading the file, take the stuff we need and store.
        if self._schema is not None:
            self._columns    = [ColumnSchema(column_details) for column_details in self._schema.get('columns', [])]
            self._column_map = ColumnMapSchema(map=self._schema.get('column_map', {}), column_names=self.ColumnNames)
        else:
            Logger.Log(f"Could not find event_data_complex schemas at {schema_path}{schema_name}", logging.ERROR)

    @property
    def ColumnNames(self) -> List[str]:
        """Function to get the names of all columns in the schema.

        :return: Names of each column in the schema.
        :rtype: List[str]
        """
        return [col.Name for col in self._columns]

    @property
    def Columns(self) -> List[ColumnSchema]:
        return self._columns

    @property
    def AsMarkdown(self) -> str:
        ret_val = "\n\n".join([
            "## Database Columns",
            "The individual columns recorded in the database for this game.",
            "\n".join([item.AsMarkdown for item in self.Columns]),
            "## Event Object Elements",
            "The elements (member variables) of each Event object, available to programmers when writing feature extractors. The right-hand side shows which database column(s) are mapped to a given element.",
            self._column_map.AsMarkdown,
            ""])
        return ret_val

    @property
    def SessionIDColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.SessionID, int):
            ret_val = self.ColumnNames[self._column_map.SessionID]
        elif isinstance(self._column_map.SessionID, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.SessionID])
        return ret_val

    @property
    def AppIDColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.AppID, int):
            ret_val = self.ColumnNames[self._column_map.AppID]
        elif isinstance(self._column_map.AppID, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.AppID])
        return ret_val

    @property
    def TimestampColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.Timestamp, int):
            ret_val = self.ColumnNames[self._column_map.Timestamp]
        elif isinstance(self._column_map.Timestamp, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.Timestamp])
        return ret_val

    @property
    def EventNameColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.EventName, int):
            ret_val = self.ColumnNames[self._column_map.EventName]
        elif isinstance(self._column_map.EventName, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.EventName])
        return ret_val

    @property
    def EventDataColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.EventData, int):
            ret_val = self.ColumnNames[self._column_map.EventData]
        elif isinstance(self._column_map.EventData, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.EventData])
        return ret_val

    @property
    def EventSourceColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.EventSource, int):
            ret_val = self.ColumnNames[self._column_map.EventSource]
        elif isinstance(self._column_map.EventSource, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.EventSource])
        return ret_val

    @property
    def AppVersionColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.AppVersion, int):
            ret_val = self.ColumnNames[self._column_map.AppVersion]
        elif isinstance(self._column_map.AppVersion, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.AppVersion])
        return ret_val

    @property
    def AppBranchColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.AppBranch, int):
            ret_val = self.ColumnNames[self._column_map.AppBranch]
        elif isinstance(self._column_map.AppBranch, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.AppBranch])
        return ret_val

    @property
    def LogVersionColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.LogVersion, int):
            ret_val = self.ColumnNames[self._column_map.LogVersion]
        elif isinstance(self._column_map.LogVersion, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.LogVersion])
        return ret_val

    @property
    def TimeOffsetColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.TimeOffset, int):
            ret_val = self.ColumnNames[self._column_map.TimeOffset]
        elif isinstance(self._column_map.TimeOffset, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.TimeOffset])
        return ret_val

    @property
    def UserIDColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.UserID, int):
            ret_val = self.ColumnNames[self._column_map.UserID]
        elif isinstance(self._column_map.UserID, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.UserID])
        return ret_val

    @property
    def UserDataColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.UserData, int):
            ret_val = self.ColumnNames[self._column_map.UserData]
        elif isinstance(self._column_map.UserData, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.UserData])
        return ret_val

    @property
    def GameStateColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.GameState, int):
            ret_val = self.ColumnNames[self._column_map.GameState]
        elif isinstance(self._column_map.GameState, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.GameState])
        return ret_val

    @property
    def EventSequenceIndexColumn(self) -> Optional[str]:
        ret_val = None
        if isinstance(self._column_map.EventSequenceIndex, int):
            ret_val = self.ColumnNames[self._column_map.EventSequenceIndex]
        elif isinstance(self._column_map.EventSequenceIndex, list):
            ret_val = ", ".join([self.ColumnNames[idx] for idx in self._column_map.EventSequenceIndex])
        return ret_val

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    _conversion_warnings = []
    def RowToEvent(self, row:Tuple, concatenator:str = '.', fallbacks:utils.map={}):
        """Function to convert a row to an Event, based on the loaded schema.
        In general, columns specified in the schema's column_map are mapped to corresponding elements of the Event.
        If the column_map gave a list, rather than a single column name, the values from each column are concatenated in order with '.' character separators.
        Finally, the concatenated values (or single value) are parsed according to the type required by Event.
        One exception: For event_data, we expect to create a Dict object, so each column in the list will have its value parsed according to the type in 'columns',
            and placed into a dict mapping the original column name to the parsed value (unless the parsed value is a dict, then it is merged into the top-level dict).

        :param row: The raw row data for an event. Generally assumed to be a tuple, though in principle a list would be fine too.
        :type  row: Tuple[str]
        :param concatenator: A string to use as a separator when concatenating multiple columns into a single Event element.
        :type  concatenator: str
        :return: [description]
        :rtype: [type]
        """
        # define vars to be passed as params
        MAX_WARNINGS : Final[int] = 10
        sess_id : str
        app_id  : str
        time    : datetime
        ename   : str
        edata   : Map
        app_ver : str
        app_br  : str
        log_ver : str
        offset  : Optional[timedelta]
        uid     : Optional[str]
        udata   : Optional[Map]
        state   : Optional[Map]
        index   : Optional[int]

        # 2) Handle event_data parameter, a special case.
        #    For this case we've got to parse the json, and then fold in whatever other columns were desired.
        # 3) Assign vals to our arg vars and pass to Event ctor.
        sess_id = self._getValueFromRow(row=row, indices=self._column_map.SessionID,   concatenator=concatenator, fallback=fallbacks.get('session_id'))
        if not isinstance(sess_id, str):
            if "sess_id" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set session_id as {type(sess_id)}, but session_id should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("sess_id")
            sess_id = str(sess_id)

        app_id  = self._getValueFromRow(row=row, indices=self._column_map.AppID,       concatenator=concatenator, fallback=fallbacks.get('app_id'))
        if not isinstance(app_id, str):
            if "app_id" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set app_id as {type(app_id)}, but app_id should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("app_id")
            app_id = str(app_id)

        time   = self._getValueFromRow(row=row, indices=self._column_map.Timestamp,   concatenator=concatenator, fallback=fallbacks.get('timestamp'))
        if not isinstance(time, datetime):
            if "timestamp" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema parsed timestamp as {type(time)}, but timestamp should be a datetime", logging.WARN)
                TableSchema._conversion_warnings.append("timestamp")
            time = TableSchema._convertDateTime(time)

        ename   = self._getValueFromRow(row=row, indices=self._column_map.EventName,   concatenator=concatenator, fallback=fallbacks.get('event_name'))
        if not isinstance(ename, str):
            if "ename" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set event_name as {type(ename)}, but event_name should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("ename")
            ename = str(ename)

        datas : Dict[str, Any] = self._getValueFromRow(row=row, indices=self._column_map.EventData,   concatenator=concatenator, fallback=fallbacks.get('event_data'))

        # TODO: go bac to isostring function; need 0-padding on ms first, though
        edata   = dict(sorted(datas.items())) # Sort keys alphabetically

        esrc    = self._getValueFromRow(row=row, indices=self._column_map.EventSource, concatenator=concatenator, fallback=fallbacks.get('event_source', EventSource.GAME))
        if not isinstance(esrc, EventSource):
            if "esrc" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set event_source as {type(esrc)}, but event_source should be an EventSource", logging.WARN)
                TableSchema._conversion_warnings.append("esrc")
            esrc = EventSource.GENERATED if esrc == "GENERATED" else EventSource.GAME

        app_ver = self._getValueFromRow(row=row, indices=self._column_map.AppVersion,  concatenator=concatenator, fallback=fallbacks.get('app_version', "0"))
        if not isinstance(app_ver, str):
            if "app_ver" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set app_version as {type(app_ver)}, but app_version should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("app_ver")
            app_ver = str(app_ver)

        app_br = self._getValueFromRow(row=row, indices=self._column_map.AppBranch,  concatenator=concatenator, fallback=fallbacks.get('app_branch'))
        if not isinstance(app_br, str):
            if "app_br" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set app_branch as {type(app_br)}, but app_branch should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("app_br")
            app_br = str(app_br)

        log_ver = self._getValueFromRow(row=row, indices=self._column_map.LogVersion,  concatenator=concatenator, fallback=fallbacks.get('log_version', "0"))
        if not isinstance(log_ver, str):
            if "log_ver" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set log_version as {type(log_ver)}, but log_version should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("log_ver")
            log_ver = str(log_ver)

        offset = self._getValueFromRow(row=row, indices=self._column_map.TimeOffset,  concatenator=concatenator, fallback=fallbacks.get('time_offset'))

        uid     = self._getValueFromRow(row=row, indices=self._column_map.UserID,      concatenator=concatenator, fallback=fallbacks.get('user_id'))
        if uid is not None and not isinstance(uid, str):
            if "uid" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set user_id as {type(uid)}, but user_id should be a string", logging.WARN)
                TableSchema._conversion_warnings.append("uid")
            uid = str(uid)

        udata   = self._getValueFromRow(row=row, indices=self._column_map.UserData,    concatenator=concatenator, fallback=fallbacks.get('user_data'))

        state   = self._getValueFromRow(row=row, indices=self._column_map.GameState,   concatenator=concatenator, fallback=fallbacks.get('game_state'))

        index   = self._getValueFromRow(row=row, indices=self._column_map.EventSequenceIndex, concatenator=concatenator, fallback=fallbacks.get('event_sequence_index'))
        if index is not None and not isinstance(index, int):
            if "index" not in TableSchema._conversion_warnings:
                Logger.Log(f"{self._table_format_name} table schema set event_sequence_index as {type(index)}, but event_sequence_index should be an int", logging.WARN)
                TableSchema._conversion_warnings.append("index")
            index = int(index)

        return Event(session_id=sess_id, app_id=app_id, timestamp=time,
                     event_name=ename, event_data=edata, event_source=esrc,
                     app_version=app_ver, app_branch=app_br, log_version=log_ver,
                     time_offset=offset, user_id=uid, user_data=udata,
                     game_state=state, event_sequence_index=index)

    # *** PRIVATE STATICS ***

    @staticmethod
    def _parse(input:Any, col_schema:ColumnSchema) -> Any:
        """Applies whatever parsing is appropriate based on what type the schema said a column contained.

        :param input: _description_
        :type input: str
        :param col_schema: _description_
        :type col_schema: ColumnSchema
        :return: _description_
        :rtype: Any
        """
        if input is None:
            return None
        if input == "None" or input == "null" or input == "nan":
            return None
        elif col_schema.ValueType == 'str':
            return str(input)
        elif col_schema.ValueType == 'int':
            return int(input)
        elif col_schema.ValueType == 'float':
            return float(input)
        elif col_schema.ValueType == 'datetime':
            return input if isinstance(input, datetime) else TableSchema._convertDateTime(str(input))
        elif col_schema.ValueType == 'timedelta':
            return input if isinstance(input, timedelta) else TableSchema._convertTimedelta(str(input))
        elif col_schema.ValueType == 'json':
            try:
                if isinstance(input, dict):
                    # if input was a dict already, then just give it back. Else, try to load it from string.
                    return input
                elif isinstance(input, str):
                    if input != 'None' and input != '': # watch out for nasty corner cases.
                        return json.loads(input)
                    else:
                        return None
                else:
                    return json.loads(str(input))
            except JSONDecodeError as err:
                Logger.Log(f"Could not parse input '{input}' of type {type(input)} from column {col_schema.Name}, got the following error:\n{str(err)}", logging.WARN)
                return {}
        elif col_schema.ValueType.startswith('enum'):
            # if the column is supposed to be an enum, for now we just stick with the string.
            return str(input)

    @staticmethod
    def _convertDateTime(time_str:str) -> datetime:
        ret_val : datetime

        if time_str == "None" or time_str == "none" or time_str == "null" or time_str == "nan":
            raise ValueError(f"Got a non-timestamp value of {time_str} when converting a datetime column of an Event!")

        formats = ["%Y-%m-%dT%H:%M:%S.%f", "%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M:%S.%f", "%Y-%m-%d %H:%M:%S.%f"]

        # for fmt in formats:
        try:
            ret_val = parser.isoparse(time_str)
            # ret_val = datetime.strptime(time_str, fmt)
        except ValueError as err:
            Logger.Log(f"Could not parse time string '{time_str}', got error {err}")
            raise err
        else:
            return ret_val
        # raise ValueError(f"Could not parse timestamp {time_str}, it did not match any expected formats.")

    @staticmethod
    def _convertTimedelta(time_str:str) -> Optional[timedelta]:
        ret_val : Optional[timedelta]

        if time_str == "None" or time_str == "none" or time_str == "null" or time_str == "nan":
            return None
        elif re.fullmatch(pattern="\d+:\d+:\d+(\.\d+)?", string=time_str):
            try:
                pieces = time_str.split(':')
                seconds_pieces = pieces[2].split('.')
                ret_val = timedelta(hours=int(pieces[0]),
                                    minutes=int(pieces[1]),
                                    seconds=int(seconds_pieces[0]),
                                    milliseconds=int(seconds_pieces[1]) if len(seconds_pieces) > 1 else 0)
            except ValueError as err:
                pass
            except IndexError as err:
                pass
            else:
                return ret_val
        elif re.fullmatch(pattern="-?\d+", string=time_str):
            try:
                ret_val = timedelta(milliseconds=int(time_str))
            except ValueError as err:
                pass
            else:
                return ret_val
        raise ValueError(f"Could not parse timedelta {time_str} of type {type(time_str)}, it did not match any expected formats.")

    # *** PRIVATE METHODS ***

    def _getValueFromRow(self, row:Tuple, indices:Union[int, List[int], Dict[str, int], None], concatenator:str, fallback:Any) -> Any:
        ret_val : Any
        if indices is not None:
            if isinstance(indices, int):
                # if there's a single index, use parse to get the value it is stated to be
                # print(f"About to parse value {row[indices]} as type {self.Columns[indices]},\nFull list from row is {row},\nFull list of columns is {self.Columns},\nwith names {self.ColumnNames}")
                ret_val = TableSchema._parse(input=row[indices], col_schema=self.Columns[indices])
            elif isinstance(indices, list):
                ret_val = concatenator.join([str(row[index]) for index in indices])
            elif isinstance(indices, dict):
                ret_val = {}
                for key,column_index in indices.items():
                    if column_index > len(row):
                        Logger.Log(f"Got column index of {column_index} for column {key}, but row only has {len(row)} columns!", logging.ERROR)
                    _val = TableSchema._parse(input=row[column_index], col_schema=self._columns[column_index])
                    ret_val.update(_val if isinstance(_val, dict) else {key:_val})
        else:
            ret_val = fallback
        return ret_val



# import libraries
import logging
from datetime import datetime
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class TotalLevelTime(PerLevelFeature):
    def __init__(self, params:GeneratorParameters):
        PerLevelFeature.__init__(self, params=params)
        self._begin_times    : List[datetime] = []
        self._complete_times : List[datetime] = []


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["BEGIN.0", "COMPLETE.0"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _validateEventCountIndex(self, event: Event):
        return self.CountIndex == 1
        


    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "BEGIN.0":
            self._begin_times.append(event.Timestamp)
        elif event.EventName == "COMPLETE.0":
            self._complete_times.append(event.Timestamp)
        else:
            Logger.Log(f"AverageLevelTime received an event which was not a BEGIN or a COMPLETE!", logging.WARN)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if len(self._begin_times) < len(self._complete_times):
            Logger.Log(f"Player began level {self.CountIndex} {len(self._begin_times)} times but completed it {len(self._complete_times)}.", logging.DEBUG)
        _num_plays = min(len(self._begin_times), len(self._complete_times))
        _diffs = [(self._complete_times[i] - self._begin_times[i]).total_seconds() for i in range(_num_plays)]
        return [sum(_diffs)]

    # *** Optionally override public functions. ***



# import standard libraries
import logging
from calendar import monthrange
from datetime import datetime
from typing import Set

# import 3rd-party libraries

# import local files
from ogd.core.interfaces.DataInterface import DataInterface
from ogd.core.interfaces.MySQLInterface import MySQLInterface
from ogd.core.interfaces.BigQueryInterface import BigQueryInterface
from ogd.core.interfaces.BQFirebaseInterface import BQFirebaseInterface
from ogd.core.requests.Request import ExporterRange
from ogd.core.schemas.configs.ConfigSchema import ConfigSchema
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.utils.Logger import Logger

class OGDGenerators:
    @staticmethod
    def genDBInterface(config:ConfigSchema, game:str) -> DataInterface:
        ret_val : DataInterface
        _game_cfg = config.GameSourceMap.get(game)
        if _game_cfg is not None and _game_cfg.Source is not None:
            match (_game_cfg.Source.Type):
                case "Firebase" | "FIREBASE":
                    ret_val = BQFirebaseInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case "BigQuery" | "BIGQUERY":
                    ret_val = BigQueryInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case "MySQL" | "MYSQL":
                    ret_val = MySQLInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case _:
                    raise Exception(f"{_game_cfg.Source.Type} is not a valid DataInterface type!")
            return ret_val
        else:
            raise ValueError(f"Config for {game} was invalid or not found!")

    @staticmethod
    def genModes(with_events:bool, with_features:bool, no_session_file:bool, no_player_file:bool, no_pop_file:bool) -> Set[ExportMode]:
        ret_val = set()

        if with_events:
            ret_val.add(ExportMode.EVENTS)
            ret_val.add(ExportMode.DETECTORS)
        if with_features:
            if not no_session_file:
                ret_val.add(ExportMode.SESSION)
            if not no_player_file:
                ret_val.add(ExportMode.PLAYER)
            if not no_pop_file:
                ret_val.add(ExportMode.POPULATION)

        return ret_val
        
    # retrieve/calculate date range.
    @staticmethod
    def genDateRange(game:str, interface, monthly:bool, start_date:str, end_date:str) -> ExporterRange:
        _from: datetime
        _to  : datetime
        today     : datetime = datetime.now()
        # If we want to export all data for a given month, calculate a date range.
        if monthly:
            month: int = today.month
            year:  int = today.year
            month_year = start_date.split("/")
            month = int(month_year[0])
            year  = int(month_year[1])
            month_range = monthrange(year, month)
            days_in_month = month_range[1]
            _from = datetime(year=year, month=month, day=1, hour=0, minute=0, second=0)
            _to   = datetime(year=year, month=month, day=days_in_month, hour=23, minute=59, second=59)
            Logger.Log(f"Exporting {month}/{year} data for {game}...", logging.DEBUG)
        # Otherwise, create date range from given pair of dates.
        else:
            _from = datetime.strptime(start_date, "%m/%d/%Y") if start_date is not None else today
            _from = _from.replace(hour=0, minute=0, second=0)
            _to   = datetime.strptime(end_date, "%m/%d/%Y") if end_date is not None else _from
            _to = _to.replace(hour=23, minute=59, second=59)
            if _from > _to:
                raise ValueError(f"Invalid date range, start date of {_from} is after end date of {_to}!")
            Logger.Log(f"Exporting from {str(_from)} to {str(_to)} of data for {game}...", logging.INFO)
        return ExporterRange.FromDateRange(source=interface, date_min=_from, date_max=_to)


# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.PENGUINS.features.bases.PerRegionFeature import PerRegionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature


class RegionEnterCount(PerCountFeature):

    def __init__(self, params:GeneratorParameters, region_map:List[Dict[str, Any]]):
        super().__init__(params=params)
        self._region_map = region_map
        self._current_count : int = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["enter_region"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._current_count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._current_count]

    # *** Optionally override public functions. ***
    def _validateEventCountIndex(self, event: Event):
        ret_val : bool = False
        region_name = event.EventData.get("region_name")
        if region_name is not None:
            if region_name in self._region_map and self._region_map[region_name] == self.CountIndex:
                ret_val = True
        else:
            self.WarningMessage(f"Got invalid region data in {type(self).__name__}")

        return ret_val


# import standard libraries
import abc
import logging
from typing import Any, Dict
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class DataSourceSchema(Schema):
    def __init__(self, name:str, other_elements:Dict[str, Any]):
        self._db_type : str
        if not isinstance(other_elements, dict):
            other_elements = {}
            Logger.Log(f"For {name} Data Source config, other_elements was not a dict, defaulting to empty dict", logging.WARN)
        # Parse DB info
        if "DB_TYPE" in other_elements.keys():
            self._db_type = DataSourceSchema._parseDBType(other_elements["DB_TYPE"])
        else:
            self._db_type = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'DB_TYPE' element; defaulting to db_host={self._db_type}", logging.WARN)

        _used = {"DB_TYPE"}
        _leftovers = { key : val for key,val in other_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def Type(self) -> str:
        """The type of source indicated by the data source schema.

        This includes but is not limited to "FIREBASE", "BIGQUERY", and "MySQL"

        :return: A string describing the type of the data source
        :rtype: str
        """
        return self._db_type

    @property
    @abc.abstractmethod
    def AsConnectionInfo(self) -> str:
        pass

    @staticmethod
    def _parseDBType(db_type) -> str:
        ret_val : str
        if isinstance(db_type, str):
            ret_val = db_type
        else:
            ret_val = str(db_type)
            Logger.Log(f"Data Source DB type was unexpected type {type(db_type)}, defaulting to str(db_type)={ret_val}.", logging.WARN)
        return ret_val



# import libraries
from datetime import datetime, timedelta
import logging, warnings
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class AverageSessionTime(Feature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._play_time: timedelta = timedelta(0)
        self._session_count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["SessionDuration"]

    def _updateFromEvent(self, event:Event) -> None:
        pass

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.ExportMode == ExtractionMode.SESSION:
            try:
                self._play_time += feature.FeatureValues[0]
                self._session_count += 1
            except TypeError as err:
                self.WarningMessage(f"TotalPlayTime for player {feature.PlayerID} got non-timedelta value of {feature.FeatureValues[0]}")

    def _getFeatureValues(self) -> List[Any]:
        return [self._play_time / self._session_count]

    # *** Optionally override public functions. ***

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.PLAYER]



## import standard libraries
import enum
import json
import logging
from collections import OrderedDict
from datetime import datetime
from typing import Any, Dict, List, Optional, Set
## import local files
from ogd.core.generators.Generator import Generator
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.generators.registries.GeneratorRegistry import GeneratorRegistry
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.IterationMode import IterationMode
from ogd.core.schemas.games.AggregateSchema import AggregateSchema
from ogd.core.schemas.games.PerCountSchema import PerCountSchema
from ogd.core.utils.Logger import Logger

## @class Extractor
#  Abstract base class for game feature extractors.
#  Gives a few static functions to be used across all extractor classes,
#  and defines an interface that the SessionProcessor can use.
class ExtractorRegistry(GeneratorRegistry):
    """Class for registering features to listen for events.

    :return: [description]
    :rtype: [type]
    """

    @enum.unique
    class FeatureOrders(enum.IntEnum):
        FIRST_ORDER = 0
        SECOND_ORDER = 1

        def __str__(self):
            return self.name

    # *** BUILT-INS & PROPERTIES ***

    # Base constructor for Registry.
    def __init__(self, mode:ExtractionMode, order:int=len(FeatureOrders)):
        """Base constructor for Registry

        Just sets up mostly-empty dictionaries for use by the registry.
        _features is a list of feature orders, where each element is a map from feature names to actual Feature instances.
        _event_registry maps event names to Listener objects, which basically just say which feature(s) wants the given enent.
        _feature_registry maps feature names to Listener objects, which basically just say which 2nd-order feature(s) wants the given 1st-order feature.
        """
        super().__init__(mode=mode)
        self._features : List[OrderedDict[str, Feature]] = [OrderedDict() for i in range(order)]
        self._feature_registry: Dict[str,List[GeneratorRegistry.Listener]] = {}
        # self._features : Dict[str, OrderedDict[str, Feature]] = {
        #     "first_order" : OrderedDict(),
        #     "second_order" : OrderedDict()
        # }

    # string conversion for Extractors.
    def __str__(self) -> str:
        """string conversion for Extractors.

        Creates a list of all features in the extractor, separated by newlines.
        :return: A string with line-separated stringified features.
        :rtype: str
        """
        ret_val : List[str] = []
        for order in self._features:
            ret_val += [str(feat) for feat in order.values()]
        return '\n'.join(ret_val)

    # Alternate string conversion for Extractors, with limitable number of lines.
    def to_string(self, num_lines:Optional[int] = None) -> str:
        """Alternate string conversion for Extractors, with limitable number of lines.

        Creates a list of features in the extractor, separated by newlines.
        Optional num_lines param allows the function caller to limit number of lines in the string.
        :param num_lines: Max number of lines to include in the string.
        If None, then include all strings, defaults to None
        :type num_lines:  Optional[int], optional
        :return: A string with line-separated stringified features.
        :rtype: str
        """
        ret_val : List[str] = []
        for order in self._features:
            ret_val += [str(feat) for feat in order.values()]
        if num_lines is None:
            return '\n'.join(ret_val)
        else:
            return '\n'.join(ret_val[:num_lines])

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _register(self, extractor:Generator, iter_mode:IterationMode):
        if isinstance(extractor, Feature):
            _listener = GeneratorRegistry.Listener(name=extractor.Name, mode=iter_mode)
            _feature_deps = extractor.FeatureFilter(mode=self._mode)
            _event_deps   = extractor.EventFilter(mode=self._mode)
            # First, add feature to the _features dict.
            if len(_feature_deps) > 0:
                _feat_order = ExtractorRegistry.FeatureOrders.SECOND_ORDER.value
            else:
                _feat_order = ExtractorRegistry.FeatureOrders.FIRST_ORDER.value
            self._features[_feat_order][extractor.Name] = extractor
            # Register feature to listen for any requested first-order features.
            for _feature_dep in _feature_deps:
                if _feature_dep not in self._feature_registry.keys():
                    self._feature_registry[_feature_dep] = []
                self._feature_registry[_feature_dep].append(_listener)
            # Finally, register feature's requested events.
            if "all_events" in _event_deps:
                self._event_registry["all_events"].append(_listener)
            else:
                for event in _event_deps:
                    if event not in self._event_registry.keys():
                        self._event_registry[event] = []
                    self._event_registry[event].append(_listener)
        else:
            raise TypeError("ExtractorRegistry was given an Extractor which was not a Feature!")

    def _getExtractorNames(self) -> List[str]:
        """Implementation of abstract function to retrieve the names of all extractors currently registered.

        :return: A list of all currently-registered features.
        :rtype: List[str]
        """
        ret_val : List[str] = []
        for order in self._features:
            for feature in order.values():
                ret_val += feature.GetFeatureNames()
        return ret_val

    def _loadFromSchema(self, schema:GameSchema, loader:GeneratorLoader, overrides:Optional[List[str]]=None):
        # first, get list of what should actually be loaded.
        # TODO : move this logic as high up as possible, so that we only need to do it once for each kind of processor.
        # 1. Start with overrides, else list of enabled features in schema.
        agg_load_set : Set[AggregateSchema]
        per_load_set : Set[PerCountSchema]
        if overrides is not None:
            agg_load_set = {schema.AggregateFeatures[name] for name in overrides if name in schema.AggregateFeatures.keys()}
            per_load_set = {schema.PerCountFeatures[name]  for name in overrides if name in schema.PerCountFeatures.keys()}
        else:
            agg_load_set = {val for val in schema.EnabledFeatures(iter_modes={IterationMode.AGGREGATE}, extract_modes={self._mode}).values() if isinstance(val, AggregateSchema)}
            per_load_set = {val for val in schema.EnabledFeatures(iter_modes={IterationMode.PERCOUNT}, extract_modes={self._mode}).values() if isinstance(val, PerCountSchema)}
        # 2. For each, grab the list of feature dependencies, and add to the list of features we want to load.
        _agg_deps = set()
        for agg in agg_load_set:
            _class = loader.GetFeatureClass(feature_type=agg.TypeName)
            if _class is not None:
                _agg_deps = _agg_deps.union(set(_class.FeatureFilter(mode=self._mode)))
        agg_load_set = agg_load_set.union({schema.AggregateFeatures[agg] for agg in _agg_deps if agg in schema.AggregateFeatures.keys()})
        _per_deps = set()
        for per in per_load_set:
            _class = loader.GetFeatureClass(feature_type=per.TypeName)
            if _class is not None:
                _per_deps = _per_deps.union(set(_class.FeatureFilter(mode=self._mode)))
        per_load_set = per_load_set.union({schema.PerCountFeatures[per] for per in _per_deps if per in schema.PerCountFeatures.keys()})
        # 3. Now that we know what needs loading, load them and register.
        for agg_schema in sorted(agg_load_set, key=lambda x : x.Name):
            feature = loader.LoadFeature(feature_type=agg_schema.TypeName, name=agg_schema.Name, schema_args=agg_schema.NonStandardElements)
            if feature is not None and self._mode in feature.AvailableModes():
                    self.Register(extractor=feature, iter_mode=IterationMode.AGGREGATE)
        for per_schema in sorted(per_load_set, key=lambda x : x.Name):
            for i in schema.GetCountRange(count=per_schema.Count):
                instance_name = f"{per_schema.Prefix}{i}_{per_schema.Name}"
                feature = loader.LoadFeature(feature_type=per_schema.TypeName, name=instance_name, schema_args=per_schema.NonStandardElements, count_index=i)
                if feature is not None and self._mode in feature.AvailableModes():
                        self.Register(extractor=feature, iter_mode=IterationMode.PERCOUNT)

    def _updateFromEvent(self, event:Event) -> None:
        """Perform extraction of features from a row.

        :param event: [description]
        :type event: Event
        :param table_schema: A data structure containing information on how the db
                             table assiciated with this game is structured.
        :type table_schema: TableSchema
        """
        listener : GeneratorRegistry.Listener = GeneratorRegistry.Listener("EMPTY", IterationMode.AGGREGATE)
        try:
            # send event to every listener for the given event name.
            for listener in self._event_registry.get(event.EventName, []):
                for order_key in range(len(self._features)):
                    if listener.name in self._features[order_key].keys():
                        self._features[order_key][listener.name].UpdateFromEvent(event)
            # don't forget to send to any features listening for "all" events
            for listener in self._event_registry["all_events"]:
                for order_key in range(len(self._features)):
                    if listener.name in self._features[order_key].keys():
                        self._features[order_key][listener.name].UpdateFromEvent(event)
        except KeyError as err:
            Logger.Log(f"{listener.name} found event {event} missing expected key: {err}", logging.ERROR)

    def _updateFromFeatureData(self, feature:FeatureData) -> None:
        """Perform extraction of features from a row.

        :param event: [description]
        :type event: Event
        :param table_schema: A data structure containing information on how the db
                             table assiciated with this game is structured.
        :type table_schema: TableSchema
        """
        listeners = self._feature_registry.get(feature.FeatureType, [])
        # send feature to every listener for the given feature name.
        for listener in listeners:
            for order_key in range(len(self._features)):
                if listener.name in self._features[order_key].keys():
                    _extractor = self._features[order_key][listener.name]
                    if feature.ExportMode in _extractor.FeatureDependencyModes():
                        self._features[order_key][listener.name].UpdateFromFeatureData(feature)


    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def OrderCount(self) -> int:
        """Gets the number of "orders" of features stored in the ExtractorRegistry.
        For now, there's just two of them.

        :return: _description_
        :rtype: int
        """
        return len(self._features)

    def GetFeatureData(self, order:int, player_id:Optional[str]=None, sess_id:Optional[str]=None) -> List[FeatureData]:
        order_index = order - 1 # orders are counted from 1, so need to adjust to index from 0.
        ret_val : List[FeatureData] = []
        for feature in self._features[order_index].values():
            ret_val.append(feature.ToFeatureData(player_id=player_id, sess_id=sess_id))
        return ret_val

    def GetFeatureValues(self) -> List[Any]:
        ret_val : List[Any] = []
        for order in self._features:
            for feature in order.values():
                next_vals = feature.GetFeatureValues()
                if len(next_vals) != len(feature.GetFeatureNames()):
                    raise ValueError(f"Feature {feature.Name} lists {len(feature.GetFeatureNames())} feature names, but returns {len(next_vals)} values!")
                ret_val += next_vals if next_vals != [] else [None]
        return ret_val

    def GetFeatureStringValues(self) -> List[str]:
        ret_val : List[str] = []
        _vals   : List[Any] = self.GetFeatureValues()

        for val in _vals:
            str_val : str
            if type(val) == dict:
                str_val = json.dumps(val)
            elif type(val) == datetime:
                str_val = val.isoformat()
            else:
                str_val = str(val)
            ret_val.append(str_val)
        return ret_val

    @property
    def FirstOrdersRequested(self) -> List[str]:
        return list(self._feature_registry.keys())

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    # def _format(obj):
    #     if obj == None:
    #         return ""
    #     elif type(obj) is timedelta:
    #         total_secs = obj.total_seconds()
    #         # h = total_secs // 3600
    #         # m = (total_secs % 3600) // 60
    #         # s = (total_secs % 3600) % 60 // 1  # just for reference
    #         # return f"{h:02.0f}:{m:02.0f}:{s:02.3f}"
    #         return str(total_secs)
    #     if obj is None:
    #         return ''
    #     else:
    #         return str(obj)



## import standard libraries
import logging
from typing import Dict, List, Set, Union

# import local files
from ogd.core.interfaces.outerfaces.DataOuterface import DataOuterface
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

class DictionaryOuterface(DataOuterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, export_modes:Set[ExportMode], out_dict:Dict[str, Dict[str, Union[List[str], List[ExportRow]]]]):
        """Constructor for a DictionaryOuterface, which provides a dictionary for each kind of data being processed

        :param game_id: The name of the game whose data is being exported
        :type game_id: str
        :param config: A GameSourceSchema indicating where output data should be stored. Ignored by the DictionaryOuterface class.
        :type config: GameSourceSchema
        :param export_modes: A set of all export modes that should be enabled.
        :type export_modes: Set[ExportMode]
        :param out_dict: The dictionary to which outputs are written by the DictionaryOuterface
        :type out_dict: Dict[str, Dict[str, Union[List[str], List[ExportRow]]]]
        """
        super().__init__(game_id=game_id, config=config, export_modes=export_modes)
        self._out = out_dict
        self._raw_evts : List[ExportRow] = []
        self._all_evts : List[ExportRow] = []
        self._sess     : List[ExportRow] = []
        self._plrs     : List[ExportRow] = []
        self._pops     : List[ExportRow] = []
        # self.Open()

    def __del__(self):
        self.Close()

    # *** IMPLEMENT ABSTRACTS ***

    def _open(self) -> bool:
        self._out['raw_events']  = { "cols" : [], "vals" : self._raw_evts }
        self._out['all_events']  = { "cols" : [], "vals" : self._all_evts }
        self._out['sessions']    = { "cols" : [], "vals" : self._sess }
        self._out['players']     = { "cols" : [], "vals" : self._plrs }
        self._out['populations'] = { "cols" : [], "vals" : self._pops }
        return True

    def _close(self) -> bool:
        return True

    def _destination(self, mode:ExportMode) -> str:
        return "RequestResult"

    def _removeExportMode(self, mode:ExportMode):
        match mode:
            case ExportMode.EVENTS:
                self._raw_evts = []
                self._out['raw_events']  = { "cols" : [], "vals" : self._raw_evts }
            case ExportMode.DETECTORS:
                self._all_evts = []
                self._out['all_events']  = { "cols" : [], "vals" : self._all_evts }
            case ExportMode.SESSION:
                self._sess = []
                self._out['sessions']    = { "cols" : [], "vals" : self._sess }
            case ExportMode.PLAYER:
                self._plrs = []
                self._out['players']     = { "cols" : [], "vals" : self._plrs }
            case ExportMode.POPULATION:
                self._pops = []
                self._out['populations'] = { "cols" : [], "vals" : self._pops }

    def _writeRawEventsHeader(self, header:List[str]) -> None:
        self._out['raw_events']['cols'] = header

    def _writeProcessedEventsHeader(self, header:List[str]) -> None:
        self._out['all_events']['cols'] = header

    def _writeSessionHeader(self, header:List[str]) -> None:
        self._out['sessions']['cols'] = header

    def _writePlayerHeader(self, header:List[str]) -> None:
        self._out['players']['cols'] = header

    def _writePopulationHeader(self, header:List[str]) -> None:
        self._out['populations']['cols'] = header

    def _writeRawEventLines(self, events:List[ExportRow]) -> None:
        # I'm always a bit fuzzy on when Python will copy vs. store reference,
        # but tests indicate if we just update self._evts, self._out is updated automatically
        # since it maps to self._evts.
        # Similar for the other functions here.
        self._raw_evts += events

    def _writeProcessedEventLines(self, events:List[ExportRow]) -> None:
        # I'm always a bit fuzzy on when Python will copy vs. store reference,
        # but tests indicate if we just update self._evts, self._out is updated automatically
        # since it maps to self._evts.
        # Similar for the other functions here.
        self._all_evts += events

    def _writeSessionLines(self, sessions:List[ExportRow]) -> None:
        self._sess += sessions

    def _writePlayerLines(self, players:List[ExportRow]) -> None:
        self._plrs += players

    def _writePopulationLines(self, populations:List[ExportRow]) -> None:
        self._pops += populations

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



from collections import defaultdict
from typing import Any, Dict, List, Union
from unicodedata import numeric

from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class PopulationSummary(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_completions = defaultdict(list)
        self._session_times = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["JobsCompleted", "SessionDuration"]

    def _updateFromEvent(self, event: Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.FeatureType == "JobsCompleted":
            self._session_completions[feature.SessionID] = feature.FeatureValues[0]
        elif feature.FeatureType == "SessionDuration":
            self._session_times.append(feature.FeatureValues[0])

    def _getFeatureValues(self) -> List[Any]:
        num_completions = [len(self._session_completions[session]) for session in self._session_completions]

        summary : Dict[str, Union[float, int]] = {
            "avg_session_count": 1,
            "avg_jobs_completed": 0,
            "avg_session_time": 0
        }

        if len(num_completions) > 0:
            summary["avg_jobs_completed"] = sum(num_completions) / len(num_completions)

        if len(self._session_times) > 0:
            summary["avg_session_time"] = sum(self._session_times) / len(self._session_times)

        return [summary]

    # *** Optionally override public functions. ***



# import libraries
from datetime import timedelta
from typing import Any, Final, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

TANK_TYPE : Final[List[str]] = [
    "Observation",
    "Stress",
    "Measurement"]

class TankRulesCount(PerCountFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._count = 0
        self._found = False

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["end_experiment", "scene_changed", "begin_experiment", "receive_fact"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "end_experiment" or event.EventName == "scene_changed":
            self._found = False
            return
        if not self._found:
            if event.EventName == "begin_experiment" and event.EventData.get("tank_type") == TANK_TYPE[self.CountIndex]:
                self._found = True
            return
        if event.EventName == "receive_fact":
            self._count += 1
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._count]

    # *** Optionally override public functions. ***
    def _validateEventCountIndex(self, event: Event):
        return True



# import standard libraries
import logging
from pathlib import Path
from typing import Any, Dict, Optional
# import local files
from ogd.core.schemas.configs.data_sources.MySQLSourceSchema import SSHSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class LegacyConfigSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._data_dir   : Optional[Path]
        self._ssh_config : Optional[SSHSchema]

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} base config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "DATA_DIR" in all_elements.keys():
            self._data_dir = LegacyConfigSchema._parseDataDir(all_elements["DATA_DIR"])
            Logger.Log(f"Found DATA_DIR legacy item in config file.", logging.INFO)
        else:
            self._data_dir = Path("./data/")
        if "SSH_CONFIG" in all_elements.keys():
            self._ssh_config = LegacyConfigSchema._parseSSHConfig(all_elements["SSH_CONFIG"])
            Logger.Log(f"Found SSH_CONFIG legacy item in config file.", logging.INFO)
        else:
            self._ssh_config = None
# 
        # _used = {"DATA_DIR", "SSH_CONFIG"}
        # _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements={})

    @property
    def DataDirectory(self) -> Optional[Path]:
        return self._data_dir

    @property
    def SSHConfig(self) -> Optional[SSHSchema]:
        return self._ssh_config

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name}"
        return ret_val

    @staticmethod
    def _parseDataDir(dir) -> Path:
        ret_val : Path
        if isinstance(dir, Path):
            ret_val = dir
        elif isinstance(dir, str):
            ret_val = Path(dir)
        else:
            ret_val = Path(str(dir))
            Logger.Log(f"Config data dir was unexpected type {type(dir)}, defaulting to Path(str(dir))={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSSHConfig(ssh) -> SSHSchema:
        ret_val : SSHSchema
        if isinstance(ssh, dict):
            ret_val = SSHSchema(name="SSH_CONFIG", all_elements=ssh)
        else:
            ret_val = SSHSchema(name="SSH_CONFIG", all_elements={})
            Logger.Log(f"Config ssh config was unexpected type {type(ssh)}, defaulting to default ssh config: {ret_val.AsMarkdown}.", logging.WARN)
        return ret_val


# import standard libraries
import itertools
import json
from pathlib import Path
from typing import Any, Callable, Dict, Final, List, Optional
# import local files
from . import features
from ogd.games import AQUALAB
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.AQUALAB.detectors import *
from ogd.games.AQUALAB.features import *
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.utils import loadJSONFile

EXPORT_PATH : Final[str] = "games/AQUALAB/DBExport.json"

## @class AqualabLoader
#  Extractor subclass for extracting features from Aqualab game data.
class AqualabLoader(GeneratorLoader):

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the AqualabLoader class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the AqualabLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        self._job_map = {"no-active-job": 0}
        self._diff_map = {0: {"experimentation": 0, "modeling": 0, "argumentation": 0} }
        self._task_map = {}
        data = None

        # Load Aqualab jobs export and map job names to integer values
        _dbexport_path = Path(AQUALAB.__file__) if Path(AQUALAB.__file__).is_dir() else Path(AQUALAB.__file__).parent
        with open(_dbexport_path / "DBExport.json", "r") as file:
            export = json.load(file)

            task_num = 1
            for i, job in enumerate(export["jobs"], start=1):
                self._job_map[job["id"]] = i
                self._diff_map[i] = job["difficulties"]
                for task in job["tasks"]:
                    task_by_job = job["id"] + "_" + task["id"]
                    self._task_map[task_by_job] = task_num
                    task_num += 1

        # Update level count
        self._game_schema._max_level = len(self._job_map) - 1

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        # First run through aggregate features
        if extractor_params._count_index == None:
            match feature_type:
                case "ActiveTime":
                    ret_val = ActiveTime.ActiveTime(params=extractor_params, job_map=self._job_map, active_threads=schema_args.get("Active_threshold"))
                case "ActiveJobs":
                    ret_val = ActiveJobs.ActiveJobs(params=extractor_params, job_map=self._job_map)
                case "AppVersions":
                    ret_val = AppVersions.AppVersions(params=extractor_params)
                case "EchoSessionID":
                    ret_val = EchoSessionID.EchoSessionID(params=extractor_params)
                case "EventList":
                    ret_val = EventList.EventList(params=extractor_params)
                case "JobsCompleted":
                    ret_val = JobsCompleted.JobsCompleted(params=extractor_params, player_id=self._player_id)
                case "JobTriesInArgument":
                    ret_val = JobTriesInArgument.JobTriesInArgument(params=extractor_params, job_map=self._job_map)
                case "ModelExportCount":
                    ret_val = ModelExportCount.ModelExportCount(params=extractor_params, job_map=self._job_map)
                case "ModelInterveneCount":
                    ret_val = ModelInterveneCount.ModelInterveneCount(params=extractor_params, job_map=self._job_map)
                case "ModelPredictCount":
                    ret_val = ModelPredictCount.ModelPredictCount(params=extractor_params, job_map=self._job_map)
                # case "PlayLocations":
                #     ret_val = PlayLocations.PlayLocations(params=extractor_params)
                case "PlayerSummary":
                    ret_val = PlayerSummary.PlayerSummary(params=extractor_params)
                case "PopulationSummary":
                    ret_val = PopulationSummary.PopulationSummary(params=extractor_params)
                case "SessionDiveSitesCount":
                    ret_val = SessionDiveSitesCount.SessionDiveSitesCount(params=extractor_params)
                case "SessionDuration":
                    ret_val = SessionDuration.SessionDuration(params=extractor_params, threshold=int(schema_args.get("threshold", 60)))
                case "SessionGuideCount":
                    ret_val = SessionGuideCount.SessionGuideCount(params=extractor_params)
                case "SessionHelpCount":
                    ret_val = SessionHelpCount.SessionHelpCount(params=extractor_params)
                case "SessionID":
                    ret_val = SessionID.SessionID(params=extractor_params, session_id=self._session_id)
                case "SessionJobsCompleted":
                    ret_val = SessionJobsCompleted.SessionJobsCompleted(params=extractor_params)
                case "SwitchJobsCount":
                    ret_val = SwitchJobsCount.SwitchJobsCount(params=extractor_params)
                case "TankRulesCount":
                    ret_val = TankRulesCount.TankRulesCount(params=extractor_params)
                case "TopJobCompletionDestinations":
                    ret_val = TopJobCompletionDestinations.TopJobCompletionDestinations(params=extractor_params, job_map=self._job_map)
                case "TopJobSwitchDestinations":
                    ret_val = TopJobSwitchDestinations.TopJobSwitchDestinations(params=extractor_params, job_map=self._job_map)
                case "TotalArgumentationTime":
                    ret_val = TotalArgumentationTime.TotalArgumentationTime(params=extractor_params)
                case "TotalDiveTime":
                    ret_val = TotalDiveTime.TotalDiveTime(params=extractor_params)
                case "TotalExperimentationTime":
                    ret_val = TotalExperimentationTime.TotalExperimentationTime(params=extractor_params)
                case "TotalGuideCount":
                    ret_val = TotalGuideCount.TotalGuideCount(params=extractor_params)
                case "TotalHelpCount":
                    ret_val = TotalHelpCount.TotalHelpCount(params=extractor_params)
                case "TotalModelingTime":
                    ret_val = TotalModelingTime.TotalModelingTime(params=extractor_params)
                case "TotalPlayTime":
                    ret_val = TotalPlayTime.TotalPlayTime(params=extractor_params)
                case "UserAvgActiveTime":
                    ret_val = UserAvgActiveTime.UserAvgActiveTime(params=extractor_params, player_id=self._player_id)
                case "UserAvgSessionDuration":
                    ret_val = UserAvgSessionDuration.UserAvgSessionDuration(params=extractor_params, player_id=self._player_id)
                case "UserTotalSessionDuration":
                    ret_val = UserTotalSessionDuration.UserTotalSessionDuration(params=extractor_params, player_id=self._player_id)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid aggregate feature type for Aqualab.")
        # then run through per-count features.
        else:
            match feature_type:
                case "JobActiveTime":
                    ret_val = JobActiveTime.JobActiveTime(params=extractor_params, job_map=self._job_map)
                case "JobArgumentation":
                    ret_val = JobArgumentation.JobArgumentation(params=extractor_params, job_map=self._job_map)
                case "JobCompletionTime":
                    ret_val = JobCompletionTime.JobCompletionTime(params=extractor_params, job_map=self._job_map)
                case "JobDiveSitesCount":
                    ret_val = JobDiveSitesCount.JobDiveSitesCount(params=extractor_params, job_map=self._job_map)
                case "JobDiveTime":
                    ret_val = JobDiveTime.JobDiveTime(params=extractor_params, job_map=self._job_map)
                case "JobExperimentation":
                    ret_val = JobExperimentation.JobExperimentation(params=extractor_params, job_map=self._job_map)
                case "JobGuideCount":
                    ret_val = JobGuideCount.JobGuideCount(params=extractor_params, job_map=self._job_map)
                case "JobHelpCount":
                    ret_val = JobHelpCount.JobHelpCount(params=extractor_params, job_map=self._job_map)
                case "JobLocationChanges":
                    ret_val = JobLocationChanges.JobLocationChanges(params=extractor_params, job_map=self._job_map)
                case "JobModeling":
                    ret_val = JobModeling.JobModeling(params=extractor_params, job_map=self._job_map)
                case "JobPriorAttempt":
                    ret_val = JobPriorAttempt.JobPriorAttempt(params=extractor_params, job_map=self._job_map)
                case "JobPriorComplete":
                    ret_val = JobPriorComplete.JobPriorComplete(params=extractor_params, job_map=self._job_map)
                case "JobStartCount":
                    ret_val = JobStartCount.JobStartCount(params=extractor_params, job_map=self._job_map)
                case "JobTasksCompleted":
                    ret_val = JobTasksCompleted.JobTasksCompleted(params=extractor_params, job_map=self._job_map)
                case "JobsAttempted":
                    ret_val = JobsAttempted.JobsAttempted(params=extractor_params, job_map=self._job_map, diff_map=self._diff_map)
                case "RegionName":
                    ret_val = RegionName.RegionName(params=extractor_params)
                case "RegionJobCount":
                    ret_val = RegionJobCount.RegionJobCount(params=extractor_params)
                case "SyncCompletionTime":
                    ret_val = SyncCompletionTime.SyncCompletionTime(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid per-count feature type for Aqualab.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        ret_val : Detector
        match detector_type:
            case "CollectFactNoJob":
                ret_val = CollectFactNoJob.CollectFactNoJob(params=extractor_params, trigger_callback=trigger_callback)
            case "DiveSiteNoEvidence":
                ret_val = DiveSiteNoEvidence.DiveSiteNoEvidence(params=extractor_params, trigger_callback=trigger_callback, threshold=schema_args['threshold'])
            case "EchoRoomChange":
                ret_val = EchoRoomChange.EchoRoomChange(params=extractor_params, trigger_callback=trigger_callback)
            case "Idle":
                ret_val = Idle.Idle(params=extractor_params, trigger_callback=trigger_callback, idle_level=schema_args.get("idle_level"))
            case "SceneChangeFrequently":
                ret_val = SceneChangeFrequently.SceneChangeFrequently(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case "HintAndLeave":
                ret_val = HintAndLeave.HintAndLeave(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case "TwoHints":
                ret_val = TwoHints.TwoHints(params=extractor_params, trigger_callback=trigger_callback, time_threshold=schema_args.get("threshold"))
            case _:
                raise NotImplementedError(f"'{detector_type}' is not a valid detector for Aqualab.")
        return ret_val

    @property
    def JobMap(self) -> Dict:
        return self._job_map

    @staticmethod
    def GetAqualabJobCount(db_export_path:Path=Path(".") / "ogd" / "games" / "AQUALAB"):
        db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
        return len(db_export.get("jobs", []))

    @staticmethod
    def GetAqualabTaskCount(db_export_path:Path=Path(".") / "ogd" / "games" / "AQUALAB"):
        db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
        list_o_lists = [job.get('tasks', []) for job in db_export.get('jobs', [])]
        # jobs_to_task_cts = [f"{job.get('id')}: {len(job.get('tasks', []))}" for job in db_export.get('jobs', [])]
        # Logger.Log(f"Task counts by job:\n{jobs_to_task_cts}", logging.DEBUG)
        all_tasks    = list(itertools.chain.from_iterable(list_o_lists))
        return len(all_tasks)



# import libraries
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event
from .. import Jowilder_Enumerators as je



# NOTE: Assumption is that in question 10 and qustion 16, we will have hover events which indicates new answer

class QuestionAnswers(PerCountFeature):
    def __init__(self, params: GeneratorParameters):
        PerCountFeature.__init__(self, params=params)
        self._guess_list : List[str] = []
        self.cur_question : Optional[int] = None
        self.chosen_answer : Optional[str] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event: Event):
        if event.EventName == "CUSTOM.20":
            return self.cur_question == self.CountIndex and self.cur_question in [10, 16]
        if Event.CompareVersions(event.LogVersion, "6") >= 0 and event.EventData.get("cur_cmd_type") == 2:
                self.cur_question = je.answer_to_question(event.EventData.get("cur_cmd_fqid"), event.GameState["level"])
        return self.cur_question==self.CountIndex
    
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.11", "CUSTOM.20"]
        # ["CUSTOM.11", "CUSTOM.20"] = ["wildcard_clicks", "wildcard_hover"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # XXX: Need a new feature as dependency to become independent from wildcard_hover event. For now, it is weird that we need to seperate two kinds of events
        # NOTE: self.chosen_answer's change: None ->interacted_fqid(click and type==2) -> interacted_fqid(hover, optional step) -> none(click and type==1, that's where we update numguesses if chosen_answer is not None)
        # NOTE: self.cur_question's function: 0 -> cur_cmd_fqid&level (cmd_type==2) -> if in [10, 16] in a wildcard_hover event, update chosen answer
        if event.event_name == "CUSTOM.20" and self.chosen_answer and event.EventData.get("cur_cmd_fqid"):
            self.chosen_answer = event.EventData.get("interacted_fqid")
            return
        if Event.CompareVersions(event.LogVersion, "6") >= 0: 
            if event.EventData.get("cur_cmd_type") == 2:
                self.cur_question = je.answer_to_question(
                    event.EventData.get("cur_cmd_fqid"), event.GameState["level"])
                self.chosen_answer = event.EventData.get("interacted_fqid")
            elif event.EventData.get("cur_cmd_type") == 1:
                if self.chosen_answer:
                    self._guess_list.append(self.chosen_answer)
                    self.chosen_answer = None
        return 

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._guess_list, len(self._guess_list)]

    def Subfeatures(self) -> List[str]:
        return ["Count"]



## import standard libraries
from datetime import datetime
import re
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.games.JOWILDER.features.UsedContinue import UsedContinue
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.games.JOWILDER.features import *



class JowilderLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        match feature_type:
            case "QuestionAnswers":
                ret_val = QuestionAnswers.QuestionAnswers(params=extractor_params)
            case "SurveyItem":
                ret_val = SurveyItem.SurveyItem(params=extractor_params)
            case "Interaction":
                ret_val = Interaction.Interaction(params=extractor_params)
            case "SurveyTime":
                ret_val = SurveyTime.SurveyTime(params=extractor_params)
            case "Clicks":
                ret_val = Clicks.Clicks(params=extractor_params)
            case "Hovers":
                ret_val = Hovers.Hovers(params=extractor_params)
            case "SessionDuration":
                ret_val = SessionDuration.SessionDuration(params=extractor_params)
            case "InteractionName":
                ret_val = InteractionName.InteractionName(params=extractor_params)
            case "NotebookUses":
                ret_val = NotebookUses.NotebookUses(params=extractor_params)
            case "EventCount":
                ret_val = EventCount.EventCount(params=extractor_params)
            case "UserEnabled":
                ret_val = UserEnabled.UserEnabled(params=extractor_params)
            case "GameVersion":
                ret_val = GameVersion.GameVersion(params=extractor_params)
            case "UsedSaveCode":
                ret_val = UsedSaveCode.UsedSaveCode(params=extractor_params)
            case "GameScript":
                ret_val = GameScript.GameScript(params=extractor_params)
            case "SessionStart":
                ret_val = SessionStart.SessionStart(params=extractor_params)
            case "IdleState":
                ret_val = IdleState.IdleState(params=extractor_params, threshold=schema_args.get("IDLE_THRESH_SECONDS", IdleState.IdleState.IDLE_TIME_THRESHOLD))
            case "ActiveStateTime":
                ret_val = ActiveStateTime.ActiveStateTime(params=extractor_params, threshold=schema_args.get('ACTIVE_THRESH_SECONDS', ActiveStateTime.ActiveStateTime.ACTIVE_TIME_THRESHOLD))
            case "MeaningfulActions":
                ret_val = MeaningfulActions.MeaningfulActions(params=extractor_params)
            case "FirstInteraction":
                ret_val = FirstInteraction.FirstInteraction(params=extractor_params)
            case "LastInteraction":
                ret_val = LastInteraction.LastInteraction(params=extractor_params)
            case "UsedContinue":
                ret_val = UsedContinue.UsedContinue(params=extractor_params)
            case "InteractionWordsPerSecond":
                ret_val = InteractionWordsPerSecond.InteractionWordsPerSecond(params=extractor_params)
            case "InteractionTextBoxesPerSecond":
                ret_val = InteractionTextBoxesPerSecond.InteractionTextBoxesPerSecond(params=extractor_params)
            case _:
                raise NotImplementedError(f"'{feature_type}' is not a valid feature for Jowilder.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid feature for Lakeland.")

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the JoWilderExtractor class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the CrystalLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)



from typing import Any, Dict, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class ToolSliderTime(SessionFeature):

    def __init__(self, params: GeneratorParameters):
        self._slider_start_time = {}
        self._slider_times = {}
        super().__init__(params=params)

    @classmethod
    def _getEventDependencies(cls, mode: ExtractionMode) -> List[str]:
        return ["grab_tool_slider", "release_tool_slider"]

    @classmethod
    def _getFeatureDependencies(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _extractFromEvent(self, event: Event) -> None:
        if event.EventName == "grab_tool_slider":
            tool_name = event.EventData["tool_name"]
            self._slider_start_time[tool_name] = event.Timestamp
        elif event.EventName == "release_tool_slider":
            tool_name = event.EventData["tool_name"]
            if tool_name in self._slider_start_time:
                start_time = self._slider_start_time[tool_name]
                end_time = event.Timestamp
                slider_time = end_time - start_time
                if tool_name in self._slider_times:
                    self._slider_times[tool_name].append(slider_time)
                else:
                    self._slider_times[tool_name] = [slider_time]

    def _extractFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._slider_times]



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature


class ActivityCompleted(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._activ_lst = []
        self._object_id = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [ "activity_end"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        self._object_id = event.event_data.get("activity_name")
        self._activ_lst.append(self._object_id)
        #if self._object_id not in self._activ_dict.keys():
            #self._activ_dict[self._object_id]=0
        #else:
            #self._activ_dict[self._object_id]+=1

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._activ_lst]





# import standard libraries
import logging
from typing import Any, Dict, List, Optional, Union
# import local files
from ogd.core.schemas.configs.data_sources.DataSourceSchema import DataSourceSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class GameSourceSchema(Schema):
    """A simple Schema structure containing configuration information for a particular game's data.
    
    When given to an interface, this schema is treated as the location from which to retrieve data.
    When given to an outerface, this schema is treated as the location in which to store data.
    (note that some interfaces/outerfaces, such as debugging i/o-faces, may ignore the configuration)
    Key properties of this schema are:
    - `Name` : Typically, the name of the Game whose source configuration is indicated by this schema
    - `Source` : A data source where game data is stored
    - `DatabaseName` : The name of the specific database within the source that contains this game's data
    - `TableName` : The neame of the specific table within the database holding the given game's data
    - `TableSchema` : A schema indicating the structure of the table containing the given game's data.

    :param Schema: _description_
    :type Schema: _type_
    """
    def __init__(self, name:str, all_elements:Dict[str, Any], data_sources:Dict[str, DataSourceSchema]):
        self._source_name   : str
        self._source_schema : Optional[DataSourceSchema]
        self._db_name       : str
        self._table_schema  : str
        self._table_name    : str

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Game Source config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "source" in all_elements.keys():
            self._source_name = GameSourceSchema._parseSource(all_elements["source"])
        else:
            self._source_name = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'source' element; defaulting to source_name={self._source_name}", logging.WARN)
        if self._source_name in data_sources.keys():
            self._source_schema = data_sources[self._source_name]
        else:
            self._source_schema = None
            Logger.Log(f"{name} config's 'source' name ({self._source_name}) was not found in available source schemas; defaulting to source_schema={self._source_schema}", logging.WARN)
        if "database" in all_elements.keys():
            self._db_name = GameSourceSchema._parseDBName(all_elements["database"])
        else:
            self._db_name = name
            Logger.Log(f"{name} config does not have a 'database' element; defaulting to db_name={self._db_name}", logging.WARN)
        if "table" in all_elements.keys():
            self._table_name = GameSourceSchema._parseTableName(all_elements["table"])
        else:
            self._table_name = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'table' element; defaulting to table={self._table_name}", logging.WARN)
        if "schema" in all_elements.keys():
            self._schema = GameSourceSchema._parseSchema(all_elements["schema"])
        else:
            self._schema = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'schema' element; defaulting to schema={self._schema}", logging.WARN)

        _used = {"source", "database", "table", "schema"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def SourceName(self) -> str:
        return self._source_name

    @property
    def Source(self) -> Optional[DataSourceSchema]:
        return self._source_schema

    @property
    def DatabaseName(self) -> str:
        return self._db_name

    @property
    def TableName(self) -> str:
        return self._table_name

    @property
    def TableSchema(self) -> str:
        return self._schema

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name}: _{self.TableSchema}_ format, source {self.Source.Name if self.Source else 'None'} : {self.DatabaseName}.{self.TableName}"
        return ret_val

    @staticmethod
    def EmptySchema() -> "GameSourceSchema":
        return GameSourceSchema(name="NOT FOUND", all_elements={}, data_sources={})

    @staticmethod
    def _parseSchema(schema) -> str:
        ret_val : str
        if isinstance(schema, str):
            ret_val = schema
        else:
            ret_val = str(schema)
            Logger.Log(f"Game Source schema type was unexpected type {type(schema)}, defaulting to str(schema)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseSource(source) -> str:
        ret_val : str
        if isinstance(source, str):
            ret_val = source
        else:
            ret_val = str(source)
            Logger.Log(f"Game Source source name was unexpected type {type(source)}, defaulting to str(source)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDBName(db_name) -> str:
        ret_val : str
        if isinstance(db_name, str):
            ret_val = db_name
        else:
            ret_val = str(db_name)
            Logger.Log(f"MySQL Data Source DB name was unexpected type {type(db_name)}, defaulting to str(db_name)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseTableName(table) -> str:
        ret_val : str
        if isinstance(table, str):
            ret_val = table
        else:
            ret_val = str(table)
            Logger.Log(f"Game Source table name was unexpected type {type(table)}, defaulting to str(table)={ret_val}.", logging.WARN)
        return ret_val



__all__ = [
    "Interface",
    "DataInterface",
		"BigQueryInterface",
		"BQFirebaseInterface",
    "CSVInterface",
		"MySQLInterface"
]

from . import Interface
from . import DataInterface
from . import BigQueryInterface
from . import BQFirebaseInterface
from . import CSVInterface
from . import MySQLInterface



## import standard libraries
import bisect
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional
## import local files
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.legacy.LegacyFeature import LegacyFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

## @class MagnetExtractor
#  Extractor subclass for extracting features from Magnet game data.
class MagnetExtractor(LegacyFeature):
    ## Constructor for the MagnetExtractor class.
    #  Initializes some custom private data (not present in base class) for use
    #  when calculating some features.
    #  Sets the sessionID feature.
    #  Further, initializes all Q&A features to -1, representing unanswered questions.
    #
    #  @param session_id The id number for the session whose data is being processed
    #                    by this extractor instance.
    #  @param game_table A data structure containing information on how the db
    #                    table assiciated with this game is structured.
    #  @param game_schema A dictionary that defines how the game data itself is
    #                     structured.
    def __init__(self, params:GeneratorParameters, game_schema:GameSchema, session_id:str):
        super().__init__(params=params, game_schema=game_schema, session_id=session_id)
        # Define custom private data.
        self._game_schema : GameSchema = game_schema
        self._features.setValByName(feature_name="sessionID", new_value=session_id)

    ## Function to perform extraction of features from a row.
    #
    #  @param row_with_complex_parsed A row of game data from the db, with the
    #                                 "complex data" already parsed from JSON.
    #  @param game_table  A data structure containing information on how the db
    #                     table assiciated with this game is structured.
    def _updateFromEvent(self, event:Event):
        # put some data in local vars, for readability later.
        level = event.GameState['level']
        if level > self._game_schema._max_level:
            Logger.Log(f"Got an event with level too high, full data:\n{str(event)}")
        # Check for invalid row.
        if self.ExtractionMode == ExtractionMode.SESSION and event.SessionID != self._session_id:
            Logger.Log(f"Got an event with incorrect session id! Expected {self._session_id}, got {event.SessionID}!",
                                logging.ERROR)
        # If row is valid, process it.
        else:
            # If we haven't set persistent id, set now.
            if self._features.getValByName(feature_name="persistentSessionID") == 0:
                self._features.setValByName(feature_name="persistentSessionID", new_value=event.UserData['persistent_session_id'])
            # Ensure we have private data initialized for this level.
            if not level in self._levels:
                bisect.insort(self._levels, level)
                self._features.initLevel(level)
            # 1) record that an event of any kind occurred, for the level & session
            self._features.incValByIndex(feature_name="eventCount", index=level)
            self._features.incAggregateVal(feature_name="sessionEventCount")
            # 2) figure out what type of event we had. If CUSTOM, we'll use the event_custom sub-item.
            event_type = event.EventName.split('.')[0]
            if event_type == "CUSTOM":
                event_type = event.EventData['event_custom']
            # 3) handle cases for each type of event
            match event_type:
                case "COMPLETE":
                    self._extractFromComplete(level=level, event_data=event.EventData)
                case "DRAG_TOOL":
                    pass
                case "DRAG_POLE":
                    pass
                case "PLAYGROUND_EXIT":
                    self._extractFromPlaygroundExit(level=level, event_data=event.EventData)
                case "TUTORIAL_EXIT":
                    self._extractFromTutorialExit(level=level, event_data=event.EventData)
                case _:
                    raise Exception(f"Found an unrecognized event type: {event_type}")

    ## Function to perform calculation of aggregate features from existing
    #  per-level/per-custom-count features.
    def _calculateAggregateFeatures(self):
        # Calculate per-level averages and percentages, since we can't calculate
        # them until we know how many total events occur.
        totalScore = 0
        for level in self._levels:
            southScore = self._features.getValByIndex(feature_name="southPoleScore", index=level)
            northScore = self._features.getValByIndex(feature_name="northPoleScore", index=level)
            totalScore = totalScore + southScore + northScore
        numPlays = self._features.getValByName(feature_name="numberOfCompletePlays")
        if numPlays != 0:
            avgScore = totalScore / numPlays
        else:
            avgScore = 0
        self._features.setValByName(feature_name="averageScore", new_value=avgScore)


    ## Private function to extract features from a "COMPLETE" event.
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    #  @param event_data        Parsed JSON data from the row being processed.
    def _extractFromComplete(self, level:int, event_data:Dict[str,Any]):
        south_score = event_data["guessScore"]["southDist"]
        north_score = event_data["guessScore"]["northDist"]
        south_score_if_switched = event_data["guessScoreIfSwitched"]["northPoleToSouthGuess"]
        north_score_if_switched = event_data["guessScoreIfSwitched"]["southPoleToNorthGuess"]
        num_compasses = event_data["numCompasses"]
        iron_filings = event_data["ironFilingsUsed"]
        magnetic_film = event_data["magneticFilmUsed"]
        times_poles_moved = event_data["numTimesPolesMoved"]
        time_spent = event_data["levelTime"]
        self._features.setValByIndex(feature_name="southPoleScore", index=level, new_value=south_score)
        self._features.setValByIndex(feature_name="northPoleScore", index=level, new_value=north_score)
        self._features.setValByIndex(feature_name="northPoleToSouthGuess", index=level, new_value=south_score_if_switched)
        self._features.setValByIndex(feature_name="southPoleToNorthGuess", index=level, new_value=north_score_if_switched)
        self._features.setValByIndex(feature_name="numberOfCompassesUsed", index=level, new_value=num_compasses)
        self._features.setValByIndex(feature_name="usedIronFilings", index=level, new_value=iron_filings)
        self._features.setValByIndex(feature_name="usedMagneticFilm", index=level, new_value=magnetic_film)
        self._features.setValByIndex(feature_name="numTimesPolesMoved", index=level, new_value=times_poles_moved)
        self._features.setValByIndex(feature_name="levelTime", index=level, new_value=time_spent)
        self._features.incAggregateVal(feature_name="sessionTime", increment=time_spent)

    ## Private function to extract features from a "PLAYGROUND_EXIT" event.
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    #  @param event_data        Parsed JSON data from the row being processed.
    def _extractFromPlaygroundExit(self, level:int, event_data:Dict[str,Any]):
        time_spent = event_data["timeSpent"]
        self._features.setValByIndex(feature_name="levelTime", index=level, new_value=time_spent)
        self._features.incAggregateVal(feature_name="sessionTime", increment=time_spent)

    ## Private function to extract features from a "TUTORIAL_EXIT" event.
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    #  @param event_data        Parsed JSON data from the row being processed.
    def _extractFromTutorialExit(self, level:int, event_data:Dict[str,Any]):
        time_spent = event_data["timeSpent"]
        self._features.setValByIndex(feature_name="levelTime", index=level, new_value=time_spent)
        self._features.incAggregateVal(feature_name="sessionTime", increment=time_spent)



# import libraries
import json
import logging
import math 
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event



class PerRegionFeature(PerCountFeature):
    def __init__(self, params:GeneratorParameters, region_map:List[Dict[str, Any]]):
        super().__init__(params=params)
        self._region_map = region_map

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event:Event):
        ret_val : bool = False
        current_position : Dict[str, float]

        int_version = int(event.LogVersion)
        match int_version:
            # Old format
            case int_version if int_version <= 9:
                current_position = {
                    'x': event.GameState.get('posX', -math.inf),
                    'y': event.GameState.get('posY', -math.inf),
                    'z': event.GameState.get('posZ', -math.inf)
                }
            # New format, up to latest version as of last change to file:
            case int_version if int_version <= 11:
                _pos = event.GameState.get("pos", [-math.inf, -math.inf, -math.inf])
                if isinstance(_pos, str):
                    try:
                        _pos = json.loads(_pos)
                    except json.JSONDecodeError as err:
                        fix_dumb_fucking_data_error = _pos[:-2] + "]"
                        _pos = json.loads(fix_dumb_fucking_data_error)
                current_position = {
                    'x': _pos[0],
                    'y': _pos[1],
                    'z': _pos[2]
                }
            # Default to current format as of last change to file:
            case _:
                _pos = event.GameState.get("pos", [-math.inf, -math.inf, -math.inf])
                if isinstance(_pos, str):
                    try:
                        _pos = json.loads(_pos)
                    except json.JSONDecodeError as err:
                        fix_dumb_fucking_data_error = _pos[:-2] + "]"
                        _pos = json.loads(fix_dumb_fucking_data_error)
                current_position = {
                    'x': _pos[0],
                    'y': _pos[1],
                    'z': _pos[2]
                }


        # region_data = event.EventData["region_name"]
        target_region = self._region_map[self.CountIndex]

        if (current_position['x'] > target_region['minX'] and 
            current_position['x'] < target_region['maxX'] and
            current_position['y'] > target_region['minY'] and
            current_position['y'] < target_region['maxY'] and
            current_position['z'] > target_region['minZ'] and
            current_position['z'] < target_region['maxZ']):
            ret_val = True
        # else:
        #     self.WarningMessage(f"Got invalid region_name data in {type(self).__name__}")
        return ret_val

        #if region_data is not None:
            #if region_data in self.region_map and self.region_map[region_data] == self.CountIndex:
                #ret_val = True
        #else:
            #self.WarningMessage(f"Got invalid region_name data in {type(self).__name__}")
        #return ret_val

    # *** Optionally override public functions. ***

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
import logging
from datetime import datetime, timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class SyncCompletionTime(Feature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None
        self._sim_start_time : Optional[datetime] = None
        self._prev_timestamp : Optional[datetime] = None
        self._time = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["begin_simulation, simulation_sync_achieved"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []
    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            _prev_id = self._session_id
            self._session_id = event.SessionID
            if self._sim_start_time is not None and self._prev_timestamp is not None:
                self._time += (self._prev_timestamp - self._sim_start_time).total_seconds()
                self._sim_start_time = event.Timestamp
            # if previous ID wasn't None, then this isn't the first session we've seen,
            # so we warn user that previous session left us without one of the things.
            elif _prev_id is not None:
                Logger.Log(f"In SyncCompletionTime, got a new session but either missing sim_start_time or prev_timestamp!", logging.WARN)

        if event.EventName == "begin_simulation":
            self._sim_start_time = event.Timestamp
        elif event.EventName == "simulation_sync_achieved":
            if self._sim_start_time is not None:
                self._time += (event.Timestamp - self._sim_start_time).total_seconds()
                self._sim_start_time = None
            else:
                self.WarningMessage("Simulation synced when we had no active start time!")

        self._prev_timestamp = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    # *** Optionally override public functions. ***



from typing import Any, List

from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class PlayerSummary(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._summary = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["JobsCompleted", "SessionDuration"]

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        session_id = feature.SessionID

        if session_id not in self._summary:
            self._summary[session_id] = {
                "active_time": 0,
                "jobs_completed": [],
                "num_sessions": 1
            }

        if feature.FeatureType == "JobsCompleted":
            self._summary[session_id]["jobs_completed"] = feature.FeatureValues[0]
        elif feature.FeatureType == "SessionDuration":
            self._summary[session_id]["active_time"] += feature.FeatureValues[0]

    def _getFeatureValues(self) -> List[Any]:
        return [self._summary]

    # *** Optionally override public functions. ***



## import standard libraries
from datetime import datetime
import re
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.games.JOURNALISM.features import *
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.games.JOURNALISM.features import StoryScoreSequence
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema



class JournalismLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        if extractor_params._count_index == None:
            match feature_type:
                case "ChoiceClickCount":
                    ret_val = ChoiceClickCount.ChoiceClickCount(params=extractor_params)
                case "SessionPlayTime":
                    ret_val = SessionPlayTime.SessionPlayTime(params=extractor_params, threshold=schema_args.get("IDLE_THRESH_SECONDS", SessionPlayTime.SessionPlayTime.IDLE_TIME_THRESHOLD))
                case "FinalAttributes":
                    ret_val = FinalAttributes.FinalAttributes(params=extractor_params)
                case "FailureAttributes":
                    ret_val = FailureAttributes.FailureAttributes(params=extractor_params)
                case "SkillSequenceCount":
                    ret_val = SkillSequenceCount.SkillSequenceCount(params = extractor_params)
                case "MeanSnippetTime":
                    ret_val = MeanSnippetTime.MeanSnippetTime(params = extractor_params)
                case "TextClickCount":
                    ret_val = TextClickCount.TextClickCount(params=extractor_params)
                case "SnippetReceivedCount":
                    ret_val = SnippetReceivedCount.SnippetReceivedCount(params =extractor_params)
                case "StoryCompleteTime":
                    ret_val = StoryCompleteTime.StoryCompleteTime(params = extractor_params)
                case "TotalLevelTime":
                    ret_val = TotalLevelTime.TotalLevelTime(params = extractor_params)
                case "PlayerAttributes":
                    ret_val = PlayerAttributes.PlayerAttributes(params = extractor_params)
                case "QuitLevel":
                    ret_val = QuitLevel.QuitLevel(params = extractor_params)
                case "QuitType":
                    ret_val = QuitType.QuitType(params = extractor_params)
                case "WorstAttribute":
                    ret_val = WorstAttribute.WorstAttribute(params = extractor_params)
                case "TopAttribute":
                    ret_val = TopAttribute.TopAttribute(params=extractor_params)
                case "TopPlayerAttribute":
                    ret_val = TopPlayerAttribute.TopPlayerAttribute(params=extractor_params)
                case "TotalFails":
                    ret_val = TotalFails.TotalFails(params=extractor_params)
                case "ContinuesOnFail":
                    ret_val = ContinuesOnFail.ContinuesOnFail(params = extractor_params)        
                case "PlayTime":
                    ret_val = PlayTime.PlayTime(params=extractor_params, threshold= schema_args.get("IDLE_THRESH_SECONDS", PlayTime.PlayTime.IDLE_TIME_THRESHOLD))
                case "UserPlayTime":
                    ret_val = UserPlayTime.UserPlayTime(params=extractor_params)
                case "GameComplete":
                    ret_val = GameComplete.GameComplete(params=extractor_params)
                case "QuitNode":
                    ret_val = QuitNode.QuitNode(params=extractor_params)
                case _:
                    raise NotImplementedError(
                        f"'{feature_type}' is not a valid aggregate feature for Journalism.")
        ##per-count features
        else:
            match feature_type:
                case "StoryAlignment":
                    ret_val = StoryAlignment.StoryAlignment(params = extractor_params)
                case "WorstPlayerAttribute":
                    ret_val = WorstPlayerAttribute.WorstPlayerAttribute(params=extractor_params)
                case "MaxedPlayerAttribute":
                    ret_val = MaxedPlayerAttribute.MaxedPlayerAttribute(params=extractor_params)           
                case "LevelCompleteCount":
                    ret_val = LevelCompleteCount.LevelCompleteCount(params=extractor_params)
                case "LevelCompleted":
                    ret_val = LevelCompleted.LevelCompleted(params=extractor_params)
                case "FailureCount":
                    ret_val = FailureCount.FailureCount(params=extractor_params)
                case "SnippetReplace":
                    ret_val = SnippetReplace.SnippetReplace(params=extractor_params)
                case "StoryEditorTime":
                    ret_val = StoryEditorTime.StoryEditorTime(params=extractor_params, threshold= schema_args.get("IDLE_THRESH_SECONDS", PlayTime.PlayTime.IDLE_TIME_THRESHOLD))
                case "AttributeView":
                    ret_val = AttributeView.AttributeView(params=extractor_params)
                case "EditorNoteOpen":
                    ret_val = EditorNoteOpen.EditorNoteOpen(params=extractor_params)
                case "TopPlayerQuitType":
                    ret_val= TopPlayerQuitType.TopPlayerQuitType(params=extractor_params)
                case "StoryScore":
                    ret_val= StoryScore.StoryScore(params=extractor_params)    
                case "StoryAlignmentSequence":
                    ret_val= StoryAlignmentSequence.StoryAlignmentSequence(params=extractor_params)    
                case "StoryScoreSequence":
                    ret_val= StoryScoreSequence.StoryScoreSequence(params=extractor_params)    
                case "LevelTime":
                    ret_val= LevelTime.LevelTime(params=extractor_params)    
                case "SnippetsCollected":
                    ret_val= SnippetsCollected.SnippetsCollected(params=extractor_params)    
                case "SnippetsSubmitted":
                    ret_val= SnippetsSubmitted.SnippetsSubmitted(params=extractor_params)
                case _:
                    raise NotImplementedError(
                        f"'{feature_type}' is not a valid per-count feature for Journalism.")
        return ret_val
    

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid detector for Journalism.")

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the JournalismExtractor class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the CrystalLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)



# import libraries
from datetime import timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class MissionSonarTimeToComplete(Feature):
    
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._sonar_start_time = None
        self._time = timedelta(0)

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["sonar_start", "sonar_exit"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "sonar_start":
            self._sonar_start_time = event.Timestamp
        elif event.EventName == "sonar_complete":
            if self._sonar_start_time is not None:
                self._time += (event.Timestamp - self._sonar_start_time).total_seconds() # TODO : maybe see if we should add timedeltas and convert to float at end?
                self._sonar_start_time = None

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._time]

    # *** Optionally override public functions. ***


# import libraries
from typing import Any, List
import json
import pandas as pd
import numpy as np
import json
import hashlib
from datetime import datetime
from datetime import timedelta
from collections import OrderedDict
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

pd.options.mode.chained_assignment = None  # default='warn'
    
listActionEvents = ['move_shape', 'rotate_shape', 'scale_shape',
                    'check_solution', 'undo_action', 'redo_action',
                    'rotate_view', 'snapshot', 'mode_change',
                    'create_shape', 'select_shape', 'delete_shape', 'select_shape_add']

orderMapping = {'1. One Box': 1, '2. Separated Boxes': 2, '3. Rotate a Pyramid': 3, '4. Match Silhouettes': 4, '5. Removing Objects': 5, '6. Stretch a Ramp': 6, '7. Max 2 Boxes': 7, '8. Combine 2 Ramps': 8, '9. Scaling Round Objects': 9,
                'Square Cross-Sections': 10, 'Bird Fez': 11, 'Pi Henge': 12, '45-Degree Rotations': 13,  'Pyramids are Strange': 14, 'Boxes Obscure Spheres': 15, 'Object Limits': 16, 'Warm Up': 17, 'Angled Silhouette': 18,
                'Sugar Cones': 19,'Stranger Shapes': 20, 'Tall and Small': 21, 'Ramp Up and Can It': 22, 'More Than Meets Your Eye': 23, 'Not Bird': 24, 'Unnecesary': 25, 'Zzz': 26, 'Bull Market': 27, 'Few Clues': 28, 'Orange Dance': 29, 'Bear Market': 30}

# mapping to positions
typeMapping = {'1. One Box': 'Basic Puzzles', '2. Separated Boxes': 'Basic Puzzles', '3. Rotate a Pyramid': 'Basic Puzzles', '4. Match Silhouettes': 'Basic Puzzles', '5. Removing Objects': 'Basic Puzzles', '6. Stretch a Ramp': 'Basic Puzzles', '7. Max 2 Boxes': 'Basic Puzzles', '8. Combine 2 Ramps': 'Basic Puzzles', '9. Scaling Round Objects': 'Basic Puzzles',
               'Square Cross-Sections': 'Intermediate Puzzles', 'Bird Fez': 'Intermediate Puzzles', 'Pi Henge': 'Intermediate Puzzles', '45-Degree Rotations': 'Intermediate Puzzles',  'Pyramids are Strange': 'Intermediate Puzzles', 'Boxes Obscure Spheres': 'Intermediate Puzzles', 'Object Limits': 'Intermediate Puzzles', 'Angled Silhouette': 'Intermediate Puzzles',
               'Sugar Cones': 'Advanced Puzzles', 'Stranger Shapes': 'Advanced Puzzles', 'Tall and Small': 'Advanced Puzzles', 'Ramp Up and Can It': 'Advanced Puzzles', 'More Than Meets Your Eye': 'Advanced Puzzles', 'Not Bird': 'Advanced Puzzles', 'Unnecessary': 'Advanced Puzzles', 'Zzz': 'Advanced Puzzles', 'Bull Market': 'Advanced Puzzles', 'Few Clues': 'Advanced Puzzles', 'Orange Dance': 'Advanced Puzzles', 'Bear Market': 'Advanced Puzzles', 'Warm Up': 'Intermediate Puzzles'}

thresHoldActivity = 60

class LevelsOfDifficulty(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._numberActions = 0
        self._numberAttempts = 0
        
        self._activePuzzle = None
        self._activeTime = []
        self._previousEvent = None
        self._userPuzzleDict = dict()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ['move_shape', 'rotate_shape', 'scale_shape',
                'check_solution', 'undo_action', 'redo_action',
                'rotate_view', 'snapshot', 'mode_change',
                'create_shape', 'select_shape', 'delete_shape',
                'deselect_shape', 'select_shape_add', 'start_level',
                'puzzle_started', 'puzzle_complete', 'restart_puzzle',
                'click_nothing', 'toggle_paint_display', 'palette_change',
                'paint', 'toggle_snapshot_display']

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        ignoreEvent = False
        if event.EventName in ["start_level", "puzzle_started"]:
            self._activePuzzle = event.EventData["task_id"]

            if self._activePuzzle not in self._userPuzzleDict.keys():
                self._userPuzzleDict[self._activePuzzle] = {"completed":0, "n_actions":0, "n_attempts":0, "active_time":0}

        elif event.EventName == "puzzle_complete":
            if self._activePuzzle in self._userPuzzleDict.keys():
                self._userPuzzleDict[self._activePuzzle]["completed"] = 1
        
        if self._activePuzzle is None:
            ignoreEvent = True
        
        if ignoreEvent == False:
            if self._previousEvent is None:
                self._previousEvent = event
                ignoreEvent = True
            
            if ignoreEvent == False:
                #Add new active_time
                delta_seconds = (event.Timestamp - self._previousEvent.timestamp).total_seconds()
                if delta_seconds < thresHoldActivity:
                    self._activeTime.append(delta_seconds)
                
                if (event.EventName in listActionEvents):
                    self._numberActions += 1
                
                if (event.EventName == "check_solution"):
                    self._numberAttempts += 1
                    
                self._previousEvent = event
                    
                if (event.EventName in ['puzzle_complete', 'exit_to_menu', 'disconnect']):
                    self._userPuzzleDict[self._activePuzzle]['n_attempts'] += self._numberAttempts
                    self._userPuzzleDict[self._activePuzzle]['n_actions'] += self._numberActions
                    self._userPuzzleDict[self._activePuzzle]['active_time'] += round(np.sum(self._activeTime)/60, 2)
                    
                    self._previousEvent = None
                    self._activeTime = []
                    self._activePuzzle = None
                    self._numberAttempts = 0
                    self._numberActions = 0
            
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        listReturn = []
        listAttempts = {}
        listActions = {}
        listActiveTime = {}
        listCompleted = {}
        listIncorrect = {}
        listAbandoned = {}
        for puzzle in self._userPuzzleDict.keys():
            listAttempts[puzzle] = self._userPuzzleDict[puzzle]['n_attempts']
            listActions[puzzle] = self._userPuzzleDict[puzzle]['n_actions']
            listActiveTime[puzzle] = self._userPuzzleDict[puzzle]['active_time']
            listCompleted[puzzle] = self._userPuzzleDict[puzzle]['completed']
            if self._userPuzzleDict[puzzle]['n_attempts'] > 0:
                listIncorrect[puzzle] = 100-100*(self._userPuzzleDict[puzzle]['completed']/self._userPuzzleDict[puzzle]['n_attempts'])
            else:
                listIncorrect[puzzle] = 100
            listAbandoned[puzzle] = 1 - self._userPuzzleDict[puzzle]['completed']
        listReturn.append("WorkInProgress")
        listReturn.append(json.dumps(listAttempts))
        listReturn.append(json.dumps(listActions))
        listReturn.append(json.dumps(listActiveTime))
        listReturn.append(json.dumps(listCompleted))
        listReturn.append(json.dumps(listIncorrect))
        listReturn.append(json.dumps(listAbandoned))
        
        return listReturn
        
    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["n_attempts", "n_actions", "active_time", "completed", "p_incorrect", "n_abandoned"]



# import libraries
from datetime import datetime, timedelta
import logging, warnings
from typing import Any, Final, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobPlayTime(Feature):
    IDLE_LEVEL : Final[int] = 30

    def __init__(self, params:GeneratorParameters, job_map:dict, active_threads:Optional[float] = None):
        self._job_map = job_map
        super().__init__(params=params)
        self._Idle_time: float = 0
        self.active_level:float = active_threads if active_threads else JobPlayTime.IDLE_LEVEL
        self._sess_duration: Optional[timedelta] = None
        self._client_start_time: Optional[datetime] = None
        self._client_end_time: Optional[datetime] = None
        

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if not self._client_start_time:
            self._client_start_time = event.Timestamp
        self._client_end_time = event.Timestamp
        
        # if event.EventName != "Idle":
        #     return

        if not self.active_level:
            self.active_level = event.EventData.get("level")
        else:
            if self.active_level != event.EventData.get("level"):
                warnings.warn("The idle event has a different threshold!")
                return
        self._Idle_time += event.EventData.get("time")
        

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self._client_start_time and self._client_end_time:
            return [self._client_end_time - self._client_start_time - timedelta(seconds=self._Idle_time)]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"


## import standard libraries
import abc
import logging
from datetime import datetime
from pprint import pformat
from typing import Any, Dict, List, Tuple, Optional, Union

# import local files
from ogd.core.interfaces.Interface import Interface
from ogd.core.models.Event import Event
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.tables.TableSchema import TableSchema
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.utils.Logger import Logger

class DataInterface(Interface):

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _allIDs(self) -> List[str]:
        pass

    @abc.abstractmethod
    def _fullDateRange(self) -> Dict[str,datetime]:
        pass

    @abc.abstractmethod
    def _rowsFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None, exclude_rows:Optional[List[str]] = None) -> List[Tuple]:
        pass

    @abc.abstractmethod
    def _IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]] = None) -> List[str]:
        pass

    @abc.abstractmethod
    def _datesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]] = None) -> Dict[str,datetime]:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, fail_fast:bool):
        super().__init__(config=config)
        self._fail_fast = fail_fast
        self._game_id : str  = game_id
        self._table_schema : TableSchema = TableSchema(schema_name=self._config.TableSchema)

    def __del__(self):
        self.Close()

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def AllIDs(self) -> Optional[List[str]]:
        ret_val = None
        if self.IsOpen():
            ret_val = self._allIDs()
        else:
            Logger.Log("Can't retrieve list of all session IDs, the source interface is not open!", logging.WARNING, depth=3)
        return ret_val

    def FullDateRange(self) -> Union[Dict[str,datetime], Dict[str,None]]:
        ret_val = {'min':None, 'max':None}
        if self.IsOpen():
            ret_val = self._fullDateRange()
        else:
            Logger.Log(f"Could not get full date range, the source interface is not open!", logging.WARNING, depth=3)
        return ret_val

    def EventsFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None, exclude_rows:Optional[List[str]]=None) -> Optional[List[Event]]:
        ret_val = None

        _curr_sess : str      = ""
        _evt_sess_index : int = 1
        if self.IsOpen():
            Logger.Log(f"Retrieving rows from IDs with {id_mode.name} ID mode.", logging.DEBUG, depth=3)
            _rows   = self._rowsFromIDs(id_list=id_list, id_mode=id_mode, versions=versions, exclude_rows=exclude_rows)
            _fallbacks = {"app_id":self._game_id}
            ret_val = []
            for row in _rows:
                try:
                    next_event = self._table_schema.RowToEvent(row=row, fallbacks=_fallbacks)
                    # in case event index was not given, we should fall back on using the order it came to us.
                    if next_event.SessionID != _curr_sess:
                        _curr_sess = next_event.SessionID
                        _evt_sess_index = 1
                    next_event.FallbackDefaults(index=_evt_sess_index)
                    _evt_sess_index += 1
                except Exception as err:
                    if self._fail_fast:
                        Logger.Log(f"Error while converting row to Event\nFull error: {err}\nRow data: {pformat(row)}", logging.ERROR, depth=2)
                        raise err
                    else:
                        Logger.Log(f"Error while converting row ({row}) to Event. This row will be skipped.\nFull error: {err}", logging.WARNING, depth=2)
                else:
                    ret_val.append(next_event)
        else:
            Logger.Log(f"Could not retrieve rows for {len(id_list)} session IDs, the source interface is not open!", logging.WARNING, depth=3)
        return ret_val

    def IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]]=None) -> Optional[List[str]]:
        ret_val = None
        if not self.IsOpen():
            str_min, str_max = min.strftime("%Y%m%d"), max.strftime("%Y%m%d")
            Logger.Log(f"Could not retrieve IDs for {str_min}-{str_max}, the source interface is not open!", logging.WARNING, depth=3)
        else:
            ret_val = self._IDsFromDates(min=min, max=max, versions=versions)
        return ret_val

    def DatesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None) -> Union[Dict[str,datetime], Dict[str,None]]:
        ret_val = {'min':None, 'max':None}
        if not self.IsOpen():
            Logger.Log(f"Could not retrieve date range {len(id_list)} session IDs, the source interface is not open!", logging.WARNING, depth=3)
        else:
            Logger.Log(f"Retrieving date range from IDs with {id_mode.name} ID mode.", logging.DEBUG, depth=3)
            ret_val = self._datesFromIDs(id_list=id_list, id_mode=id_mode, versions=versions)
        return ret_val

    # *** PROPERTIES ***

    @property
    def _TableSchema(self) -> TableSchema:
        return self._table_schema

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import abc
import logging
from importlib import import_module
from typing import Any, Callable, Dict, List, Optional, Type
# import locals
from ogd.core.generators.Generator import Generator, GeneratorParameters
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.extractors.Extractor import Extractor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger

class GeneratorLoader(abc.ABC):

    # *** ABSTRACTS ***
    
    @abc.abstractmethod
    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Extractor:
        pass
    
    @abc.abstractmethod
    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        pass

    @staticmethod
    @abc.abstractmethod
    def _getFeaturesModule():
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, player_id:str, session_id:str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Base constructor for Extractor classes.
        The constructor sets an extractor's session id and range of levels,
        as well as initializing the feature
        es dictionary and list of played levels.

        :param session_id: The id of the session from which we will extract features.
        :type session_id: str
        :param game_schema: A dictionary that defines how the game data itself is structured.
        :type game_schema: GameSchema
        """
        self._player_id   : str            = player_id
        self._session_id  : str            = session_id
        self._game_schema : GameSchema     = game_schema
        self._mode        : ExtractionMode = mode
        self._overrides   : Optional[List[str]] = feature_overrides

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***
    
    def LoadDetector(self, detector_type:str, name:str, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None], count_index:Optional[int] = None) -> Optional[Detector]:
        ret_val = None

        params = GeneratorParameters(name=name, description=schema_args.get('description',""), mode=self._mode, count_index=count_index)
        try:
            ret_val = self._loadDetector(detector_type=detector_type, extractor_params=params, schema_args=schema_args, trigger_callback=trigger_callback)
        except NotImplementedError as err:
            Logger.Log(f"In GeneratorLoader, '{name}' is not a valid detector for {self._game_schema.GameName}", logging.ERROR)

        return ret_val

    def LoadFeature(self, feature_type:str, name:str, schema_args:Dict[str,Any], count_index:Optional[int] = None) -> Optional[Extractor]:
        ret_val = None

        if self._validateMode(feature_type=feature_type):
            params = GeneratorParameters(name=name, description=schema_args.get('description',""), mode=self._mode, count_index=count_index)
            try:
                ret_val = self._loadFeature(feature_type=feature_type, extractor_params=params, schema_args=schema_args)
            except NotImplementedError as err:
                Logger.Log(f"In GeneratorLoader, unable to load '{name}', {feature_type} is not implemented!:", logging.ERROR)
                Logger.Log(str(err), logging.ERROR, depth=1)

        return ret_val

    def GetFeatureClass(self, feature_type:str) -> Optional[Type[Extractor]]:
        ret_val : Optional[Type[Extractor]] = None
        base_mod = self._getFeaturesModule()
        try:
            feature_mod = getattr(base_mod, feature_type)
            ret_val     = getattr(feature_mod, feature_type)
        except NameError as err:
            Logger.Log(f"Could not get class {feature_type}, a NameError occurred:\n{err}", logging.WARN)
        finally:
            return ret_val

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _validateMode(self, feature_type) -> bool:
        ret_val = False

        mod_name = f"ogd.games.{self._game_schema.GameName}.features.{feature_type}"
        try:
            feature_module = import_module(mod_name)
        except ModuleNotFoundError:
            Logger.Log(f"In GeneratorLoader, '{mod_name}' could not be found, skipping {feature_type}", logging.ERROR)
        else:
            # bit of a hack using globals() here, but theoretically this lets us access the class object with only a string.
            # feature_class : Optional[Type[Extractor]] = globals().get(feature_type, None)
            feature_class : Optional[Type[Extractor]] = getattr(feature_module, feature_type, None)
            if feature_class is not None:
                ret_val = self._mode in feature_class.AvailableModes()
            else:
                Logger.Log(f"In GeneratorLoader, feature class '{feature_type}' could not be found in module {feature_module}", logging.WARN)

        return ret_val



## import standard libraries
import abc
import typing
import logging
## import local libraries
from models.Model import *

## @class FeatureModel
#  Abstract base class for models that use per-session feature data as input rows.
#  For these models, we wrap the functionality in a private _eval function, which is called once per row.
class FeatureModel(Model):
    def __init__(self, levels: typing.List[int] = []):
        super().__init__(levels=levels, input_type=ModelInputType.FEATURE)
    ## Very simple implementation of the Eval function.
    #  We assume each row is a session, so call private _eval function once per row.
    #
    #  @param rows : A list of rows, where each row contains all columns expected by the specific model implementation.
    #  @return     : A list of results for the data (how many results depends on the model)
    def Eval(self, rows: typing.List) -> typing.List:
        return [self._eval(row) for row in rows]

    ## Abstract declaration of a function to perform calculation of a model results from a row.
    #
    #  @param rows : A row of data for a session, which should be a mapping of column names to 
    #  @return     : A result for the given row of data
    @abc.abstractmethod
    def _eval(self, row: typing.Dict):
        pass


# import libraries
import json
import logging
from collections import defaultdict
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class TopJobCompletionDestinations(Feature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        self._job_map = job_map
        super().__init__(params=params)
        self._current_user_code = None
        self._last_completed_id = None
        self._job_complete_pairs = defaultdict(dict)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["accept_job", "complete_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        _job_name = event.GameState.get('job_name', event.EventData.get('job_name', None))
        if _job_name is None:
            raise KeyError("Could not find key 'job_name' in GameState or EventData!")
        if self._validate_job(_job_name):
            user_code = event.UserID

            # in either case, handle event.
            if event.EventName == "complete_job":
                self._last_completed_id = _job_name # here, we take what we last completed, and append where we switched to.
            elif event.EventName == "accept_job":
                if user_code == self._current_user_code and self._last_completed_id is not None:
                    if not _job_name in self._job_complete_pairs[self._last_completed_id].keys():
                        self._job_complete_pairs[self._last_completed_id][_job_name] = []

                    self._job_complete_pairs[self._last_completed_id][_job_name].append(user_code)
                    self._last_completed_id = None

            # finally, once we process the event, we know we're looking at data for this event's user.
            self._current_user_code = user_code

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        ret_val = {}

        for src in self._job_complete_pairs.keys():
            dests = sorted(
                self._job_complete_pairs[src].items(),
                key=lambda item: len(item[1]), # sort by length of list of ids.
                reverse=True # sort largest to smallest
            )
            ret_val[src] = {
                item[0]:item[1] for item in dests[0:5]
            }

            # Logger.Log(f"For TopJobCompletionDestinations, sorted dests as: {json.dumps(dests)}")

        # TODO: figure out if we really need to dump to string, or if we can assume things get stringified as needed elsewhere.
        return [json.dumps(ret_val)]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    # *** Other local functions
    def _validate_job(self, job_data):
        ret_val : bool = False
        if job_data and job_data in self._job_map:
                ret_val = True
        else:
            self.WarningMessage(f"Got invalid job_name data in JobsAttempted")
        return ret_val



# import standard libraries
from datetime import datetime, timedelta
from time import time
from typing import Callable, Final, List, Optional, Union
# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.utils.typing import Map


class TwoHints(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    DEFAULT_THRESOLD : Final[int] = 5


    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None], time_threshold:Optional[int]):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._found = False
        self._sess_id = "Unknown"
        self._job_name = "Unknown"
        self._last_hint = "Unknown"
        self._this_hint = "Unknown"
        self._last_hint_time:Optional[datetime] = None
        self._time_spent: Optional[Union[timedelta, float]] = None
        self._detector_event_data: Map = {"job_name": self._job_name}
        self._threshold : timedelta
        if time_threshold is not None:
            self._threshold = timedelta(seconds=time_threshold)
        else:
            self._threshold = timedelta(seconds=TwoHints.DEFAULT_THRESOLD)

    # *** Implement abstract functions ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["ask_for_help"] # >>> fill in names of events this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        if self._sess_id == "Unknown":
            self._sess_id = event.SessionID
        elif self._sess_id != event.SessionID:
            self._sess_id = event.SessionID
            self._last_hint_time = None
        
        if not self._last_hint_time:
            self._last_hint_time = event.Timestamp
            self._last_hint = event.EventData.get("node_id")
            return
        
        self._time_spent = event.Timestamp - self._last_hint_time

        if self._time_spent <= self._threshold:
            self._found = True
            self._time_spent = self._time_spent / timedelta(seconds=1)
            self._sess_id = event.SessionID

            self._job_name = event.GameState.get('job_name', event.EventData.get('job_name', "JOB NAME NOT FOUND"))
            self._this_hint = event.EventData.get("node_id")
            self._detector_event_data = {
                "time": self._time_spent,
                "level": self._threshold / timedelta(seconds=1),
                "job_name": self._job_name,
                "last_hint_node": self._last_hint,
                "this_hint_node": self._this_hint
            }

        self._last_hint_time = event.Timestamp
        self._last_hint = self._this_hint
        return

    def _trigger_condition(self) -> bool:
        if self._found:
            self._found = False
            return True
        else:
            return False

    def _trigger_event(self) -> DetectorEvent:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val: DetectorEvent = self.GenerateEvent(session_id=self._sess_id, app_id="AQUALAB",
                                                    event_name="TwoHints", event_data=self._detector_event_data)
        return ret_val



# import libraries
import json
from typing import Any, List
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class EventList(Feature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._event_list = []
        self._mission_id = None

        # Map of event names to primary detail parameter and its type
        self._details_map = {
            "scene_load":               ("scene", "string_value"),
            "checkpoint":               ("status", "string_value"),
            "new_evidence":             ("evidence_id", "string_value"),
            "sonar_percentage_update":  ("percentage", "int_value"),
            "dive_moveto_location":     ("next_node_id", "string_value"),
            "dive_photo_click":         ("accurate", "string_value"),
            "view_dialog":              ("dialog_id", "string_value")
        }

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "checkpoint" and event.EventData["status"] == "Begin Mission":
            self._mission_id = event.EventData["mission_id"]

        next_event = {
            "name": event.EventName,
            "user_id": event.UserID,
            "session_id": event.SessionID,
            "timestamp": event.Timestamp.isoformat(),
            "job_name": self._mission_id,
            "index": event.EventSequenceIndex,
            "event_primary_detail": None
        }

        if event.EventName in self._details_map:
            param_name = self._details_map[event.EventName][0]

            next_event["event_primary_detail"] = event.EventData[param_name]

        self._event_list.append(next_event)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [json.dumps(self._event_list)]
